import json

from more_itertools import pairwise


def update_response_json_body(response):
    content_type = (
        response["headers"].get("Content-Type")
        or response["headers"].get("content-type")
        or response["headers"].get("CONTENT-TYPE")
    )
    response["content_type"] = content_type
    content_type_is_json = content_type and "application/json" in content_type
    if response["body"] and content_type_is_json:
        response["json_body"] = json.loads(response["body"])


def parse_outbound_frames(frames):
    frames = (
        frame
        for frame in frames
        if frame["type"] in ("outbound_http_request", "outbound_http_response")
    )

    without_urllib3 = (frame for frame in frames if frame["subtype"] != "urllib3")

    outbound_frames = []
    for request, response in pairwise(without_urllib3):
        if (
            request["type"] != "outbound_http_request"
            or response["type"] != "outbound_http_response"
        ):
            # We don't have a (request, response) pair, so move ahead.
            # Usually this is (response, request), but could also be
            # (request, request) or (response, response) if the trace
            # failed to record as expected.
            continue  # pragma: no cover

        update_response_json_body(response)
        outbound_frames.append({"request": request, "response": response})

    return outbound_frames



from __future__ import annotations

import os
import threading
import time
import types
from typing import Dict, List, Literal, Tuple

import ulid

from ..serialize import decode_body, decode_header_value
from .types import BaseOutboundRequest, BaseOutboundResponse


class OutboundRequest(BaseOutboundRequest):
    subtype: Literal["urllib"]


class OutboundResponse(BaseOutboundResponse):
    subtype: Literal["urllib"]


class UrllibFilter:
    co_names: Tuple[str, ...] = ("do_open",)
    urllib_filename = os.path.normpath("urllib/request")

    def __init__(self, config) -> None:
        self.config = config
        self._frame_ids: Dict[int, str] = {}

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        return (
            frame.f_code.co_name == "do_open"
            and self.urllib_filename in frame.f_code.co_filename
        )

    def process(
        self,
        frame: types.FrameType,
        event: str,
        arg: object,
        call_frames: List[Tuple[types.FrameType, str]],
    ):
        thread = threading.current_thread()
        request = frame.f_locals["req"]
        full_url = request.full_url
        method = request.get_method()
        method_and_full_url = f"{method} {full_url}"
        if event == "call":
            frame_id = f"frm_{ulid.new()}"
            self._frame_ids[id(frame)] = frame_id
            request_headers = {
                key: decode_header_value(value) for key, value in request.header_items()
            }

            api_request: OutboundRequest = {
                "body": decode_body(request.data, request_headers),
                "frame_id": frame_id,
                "headers": request_headers,
                "method": method,
                "method_and_full_url": method_and_full_url,
                "subtype": "urllib",
                "thread": thread.name,
                "thread_native_id": thread.native_id,
                "timestamp": time.time(),
                "type": "outbound_http_request",
                "url": full_url,
            }
            return api_request

        elif event == "return":  # pragma: no branch
            if arg is None:
                # Probably an exception
                headers = None
                status_code = None
            else:
                response = frame.f_locals["r"]
                headers = dict(response.headers)
                status_code = response.status

            api_response: OutboundResponse = {
                "body": None,
                "frame_id": self._frame_ids[id(frame)],
                "headers": headers,
                "method": method,
                "method_and_full_url": method_and_full_url,
                "status_code": status_code,
                "subtype": "urllib",
                "thread": thread.name,
                "thread_native_id": thread.native_id,
                "timestamp": time.time(),
                "type": "outbound_http_response",
                "url": full_url,
            }
            return api_response



from __future__ import annotations

import inspect
import json
import logging
import sys
import threading
import time
import types
from collections import defaultdict
from collections.abc import Mapping
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Dict, List, Protocol, TypeVar, overload

import msgpack
import ulid

from .config import load_config
from .db import save_trace_in_sqlite, setup_db
from .filters.attrs import attrs_filter
from .filters.celery import CeleryFilter
from .filters.core import (
    FrameFilter,
    FrameProcessor,
    build_frame_filter,
    exec_filter,
    frozen_filter,
    library_filter,
)
from .filters.django import (
    DjangoChecksFilter,
    DjangoFilter,
    DjangoSetupFilter,
    DjangoTemplateFilter,
    DjangoTestDBFilter,
)
from .filters.exception import ExceptionFilter
from .filters.httpx import HttpxFilter
from .filters.huey import HueyFilter
from .filters.kolo import kolo_filter
from .filters.logging import LoggingFilter
from .filters.pypy import pypy_filter
from .filters.pytest import PytestFilter
from .filters.requests import ApiRequestFilter
from .filters.sql import SQLQueryFilter
from .filters.unittest import UnitTestFilter
from .filters.urllib import UrllibFilter
from .filters.urllib3 import Urllib3Filter
from .git import COMMIT_SHA
from .serialize import (
    UserCodeCallSite,
    dump_data,
    dump_json,
    dump_msgpack,
    frame_path,
    monkeypatch_queryset_repr,
)
from .version import __version__

logger = logging.getLogger("kolo")


class KoloLocals(threading.local):
    def __init__(self):
        self.call_frames = []
        self._frame_ids = {}


class KoloProfiler:
    """
    Collect runtime information about code to view in VSCode.

    include_frames can be passed to enable profiling of standard library
    or third party code.

    ignore_frames can also be passed to disable profiling of a user's
    own code.

    The list should contain fragments of the path to the relevant files.
    For example, to include profiling for the json module the include_frames
    could look like ["/json/"].

    The list may also contain frame filters. A frame filter is a function
    (or other callable) that takes the same arguments as the profilefunc
    passed to sys.setprofile and returns a boolean representing whether
    to allow or block the frame.

    include_frames takes precedence over ignore_frames. A frame that
    matches an entry in each list will be profiled.
    """

    def __init__(
        self, db_path: Path, config=None, one_trace_per_test=False, *, source
    ) -> None:
        self.db_path = db_path
        self.source = source
        self.one_trace_per_test = one_trace_per_test
        trace_id = ulid.new()
        self.trace_id = f"trc_{trace_id}"
        self.start_test_index = 0
        self.start_test_indices: Dict[int, int] = {}
        self.frames_of_interest: List[str] = []
        self.frames: defaultdict = defaultdict(list)
        self.config = config if config is not None else {}
        filter_config = self.config.get("filters", {})
        include_frames = filter_config.get("include_frames", ())
        ignore_frames = filter_config.get("ignore_frames", ())
        self.include_frames = list(map(build_frame_filter, include_frames))
        self.ignore_frames = list(map(build_frame_filter, ignore_frames))
        # The order here matters for the Rust implementation, which accesses
        # entries by index.
        self._default_include_frames: List[FrameProcessor] = [
            DjangoFilter(self.config),
            DjangoTemplateFilter(self.config),
            CeleryFilter(self.config),
            HueyFilter(self.config),
            ApiRequestFilter(self.config),
            UrllibFilter(self.config),
            Urllib3Filter(self.config),
            ExceptionFilter(
                self.config,
                ignore_frames=self.ignore_frames,
                include_frames=self.include_frames,
            ),
            LoggingFilter(self.config),
            SQLQueryFilter(self.config),
            UnitTestFilter(self.config),
            PytestFilter(self.config),
            HttpxFilter(self.config),
            DjangoSetupFilter(self.config),
            DjangoChecksFilter(self.config),
            DjangoTestDBFilter(self.config),
        ]

        self.default_include_frames: Dict[str, List[FrameProcessor]] = {}
        for filter in self._default_include_frames:
            for co_name in filter.co_names:
                self.default_include_frames.setdefault(co_name, []).append(filter)

        self.default_ignore_frames: List[FrameFilter] = [
            library_filter,
            frozen_filter,
            pypy_filter,
            kolo_filter,
            exec_filter,
            attrs_filter,
        ]
        self.thread_locals = KoloLocals()
        self.timestamp = time.time()
        self.rust_profiler = None
        self.main_thread_id = threading.main_thread().native_id

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> None:
        if event in ["c_call", "c_return"]:
            return

        for frame_filter in self.include_frames:
            try:
                if frame_filter(frame, event, arg):
                    self.process_frame(frame, event, arg)
                    return
            except Exception as e:
                logger.warning(
                    "Unexpected exception in include_frames: %s",
                    frame_filter,
                    exc_info=e,
                )
                continue

        for frame_filter in self.ignore_frames:
            try:
                if frame_filter(frame, event, arg):
                    return
            except Exception as e:
                logger.warning(
                    "Unexpected exception in ignore_frames: %s",
                    frame_filter,
                    exc_info=e,
                )
                continue

        co_name = frame.f_code.co_name

        # Execute only the filters listening for this co_name
        for frame_filter in self.default_include_frames.get(co_name, ()):
            try:
                if frame_filter(frame, event, arg):
                    frame_data = frame_filter.process(
                        frame, event, arg, self.thread_locals.call_frames
                    )
                    if frame_data:  # pragma: no branch
                        json_data = dump_data(frame_data, self.config)
                        if (
                            self.one_trace_per_test
                            and frame_data["type"] == "start_test"
                        ):
                            self.start_test()  # pragma: no cover

                        if self.config.get("threading", False):  # pragma: no cover
                            thread = threading.current_thread()
                            if thread.native_id == self.main_thread_id:
                                self.frames_of_interest.append(json_data)
                            else:  # pragma: no cover
                                self.frames[thread.native_id].append(json_data)
                        else:
                            self.frames_of_interest.append(json_data)

                        if self.one_trace_per_test and frame_data["type"] == "end_test":
                            self.end_test()  # pragma: no cover
                    return
            except Exception as e:
                logger.warning(
                    "Unexpected exception in default_include_frames: %s",
                    frame_filter,
                    exc_info=e,
                )
                continue

        for frame_filter in self.default_ignore_frames:
            try:
                if frame_filter(frame, event, arg):
                    return
            except Exception as e:
                logger.warning(
                    "Unexpected exception in default_ignore_frames: %s",
                    frame_filter,
                    exc_info=e,
                )
                continue

        try:
            self.process_frame(frame, event, arg)
        except Exception as e:
            logger.warning(
                "Unexpected exception in KoloProfiler.process_frame",
                exc_info=e,
            )

    def start_test(self):
        self.trace_id = f"trc_{ulid.new()}"
        self.start_test_index = len(self.frames_of_interest)
        self.start_test_indices = {
            thread_id: len(frames) for thread_id, frames in self.frames.items()
        }

    def end_test(self):
        test_frames = {
            thread_id: frames[self.start_test_indices.get(thread_id, 0) :]
            for thread_id, frames in self.frames.items()
        }
        self.save_request_in_db(
            self.frames_of_interest[self.start_test_index :],
            test_frames,
        )

    def __enter__(self) -> None:
        if self.config.get("use_rust", True):
            try:
                from ._kolo import register_profiler
            except ImportError:
                sys.setprofile(self)
                if self.config.get("threading", False):
                    threading.setprofile(self)
            else:
                register_profiler(self)
        else:
            sys.setprofile(self)
            if self.config.get("threading", False):
                threading.setprofile(self)

    def __exit__(self, *exc) -> None:
        sys.setprofile(None)
        if self.config.get("threading", False):
            threading.setprofile(None)

    def format_json_data(
        self,
        data: Dict[str, Any],
        frames: List[str],
        frames_by_thread: Dict[int, List[str]],
    ) -> str:
        """
        Build a json blob from trace data and frame data

        `frames` is a list of json strings, so if we naïvely add it to `data` and
        dump it as json, we'll double encode it. Instead, we build the json array
        with some string formatting and replace the `frames_placeholder` in data
        with more string formatting.

        We handle `self.frames` similarly.
        """
        frames_placeholder = "KOLO_FRAMES_OF_INTEREST"
        frames_by_thread_placeholder = "KOLO_FRAMES_BY_THREAD"
        data["frames_of_interest"] = frames_placeholder
        data["frames"] = frames_by_thread_placeholder
        json_data = dump_json(data)

        json_frames_by_thread = ",\n".join(
            f'"{thread_id}": [{", ".join(thread_frames)}]'
            for thread_id, thread_frames in frames_by_thread.items()
        )
        json_frames = ", ".join(frames)
        return json_data.replace(
            json.dumps(frames_placeholder), f"[{json_frames}]"
        ).replace(
            json.dumps(frames_by_thread_placeholder), f"{{{json_frames_by_thread}}}"
        )

    def save_request_in_db(self, frames=None, thread_frames=None) -> None:
        if self.rust_profiler:
            self.rust_profiler.save_request_in_db()
            return

        timeout = self.config.get("sqlite_busy_timeout", 60)
        timestamp = self.timestamp
        use_msgpack = self.config.get("use_msgpack", True)
        main_thread_id = (
            self.main_thread_id if use_msgpack else str(self.main_thread_id)
        )
        data = {
            "command_line_args": sys.argv,
            "current_commit_sha": COMMIT_SHA,
            "main_thread_id": main_thread_id,
            "meta": {
                "version": __version__,
                "source": self.source,
                "use_frame_boundaries": True,
            },
            "timestamp": timestamp,
            "trace_id": self.trace_id,
        }

        frames = self.frames_of_interest if frames is None else frames
        thread_frames = self.frames if thread_frames is None else thread_frames

        if use_msgpack:
            data["frames_of_interest"] = [
                msgpack.unpackb(f, strict_map_key=False) for f in frames
            ]
            data["frames"] = {
                thread_id: [msgpack.unpackb(f, strict_map_key=False) for f in frames]
                for thread_id, frames in thread_frames.items()
            }
            serialized_data = dump_msgpack(data)
            save_trace_in_sqlite(
                self.db_path, self.trace_id, msgpack=serialized_data, timeout=timeout
            )
        else:
            serialized_data = self.format_json_data(data, frames, thread_frames)
            save_trace_in_sqlite(
                self.db_path,
                self.trace_id,
                json_string=serialized_data,
                timeout=timeout,
            )

    def process_frame(self, frame: types.FrameType, event: str, arg: object) -> None:
        user_code_call_site: UserCodeCallSite | None
        if event == "call" and self.thread_locals.call_frames:
            call_frame, call_frame_id = self.thread_locals.call_frames[-1]
            user_code_call_site = {
                "call_frame_id": call_frame_id,
                "line_number": call_frame.f_lineno,
            }
        else:
            # If we are a return frame, we don't bother duplicating
            # information for the call frame.
            # If we are the first call frame, we don't have a callsite.
            user_code_call_site = None

        co_name = frame.f_code.co_name
        if event == "call":
            frame_id = f"frm_{ulid.new()}"
            self.thread_locals._frame_ids[id(frame)] = frame_id
            self.thread_locals.call_frames.append((frame, frame_id))
        elif event == "return":  # pragma: no branch
            self.thread_locals.call_frames.pop()

        thread = threading.current_thread()

        frame_locals = {k: v for k, v in frame.f_locals.items() if k != "__builtins__"}
        frame_data = {
            "path": frame_path(frame),
            "co_name": co_name,
            "qualname": get_qualname(frame),
            "event": event,
            "frame_id": self.thread_locals._frame_ids[id(frame)],
            "arg": arg,
            "locals": frame_locals,
            "thread": thread.name,
            # The operating system's thread ID.
            # Guaranteed to be unique for the lifetime of the thread.
            "thread_native_id": thread.native_id,
            "timestamp": time.time(),
            "type": "frame",
            "user_code_call_site": user_code_call_site,
        }
        json_data = dump_data(frame_data, self.config)
        if thread.native_id == self.main_thread_id:
            self.frames_of_interest.append(json_data)
        else:  # pragma: no cover
            self.frames[thread.native_id].append(json_data)


def get_qualname(frame: types.FrameType) -> str | None:
    try:
        qualname = frame.f_code.co_qualname  # type: ignore[attr-defined]
    except AttributeError:
        pass
    else:
        module = frame.f_globals["__name__"]
        return f"{module}.{qualname}"

    co_name = frame.f_code.co_name
    if co_name == "<module>":  # pragma: no cover
        module = frame.f_globals["__name__"]
        return f"{module}.<module>"

    try:
        outer_frame = frame.f_back
        assert outer_frame
        try:
            function = outer_frame.f_locals[co_name]
        except KeyError:
            try:
                self = frame.f_locals["self"]
            except KeyError:
                cls = frame.f_locals.get("cls")
                if isinstance(cls, type):
                    function = inspect.getattr_static(cls, co_name)
                else:
                    try:
                        qualname = frame.f_locals["__qualname__"]
                    except KeyError:
                        function = frame.f_globals[co_name]
                    else:  # pragma: no cover
                        module = frame.f_globals["__name__"]
                        return f"{module}.{qualname}"
            else:
                function = inspect.getattr_static(self, co_name)
                if isinstance(function, property):
                    function = function.fget

        return f"{function.__module__}.{function.__qualname__}"
    except Exception:
        return None


class Enabled:
    def __init__(self, config: Mapping[str, Any] | None = None):
        if config is None:
            config = {}
        self.config = config
        self._profiler: KoloProfiler | None = None

    def __call__(self, func):
        @wraps(func)
        def inner(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return inner

    def __enter__(self) -> None:
        if sys.getprofile():
            return

        if self.config.get("threading", False):
            try:
                thread_profiler = threading.getprofile()  # type: ignore[attr-defined]
            except AttributeError:
                thread_profiler = threading._profile_hook
            if thread_profiler:
                return

        config = load_config(self.config)
        db_path = setup_db()
        monkeypatch_queryset_repr()
        self._profiler = KoloProfiler(db_path, config=config, source="kolo.enable")
        self._profiler.__enter__()

    def __exit__(self, *exc) -> None:
        if self._profiler is not None:
            self._profiler.__exit__(*exc)
        if self._profiler is not None:
            self._profiler.save_request_in_db()
            self._profiler = None


F = TypeVar("F", bound=Callable[..., Any])


class CallableContextManager(Protocol):
    def __call__(self, func: F) -> F:
        ...  # pragma: no cover

    def __enter__(self) -> None:
        ...  # pragma: no cover

    def __exit__(self, *exc) -> None:
        ...  # pragma: no cover


@overload
def enable(_func: F) -> F:
    """Stub"""


@overload
def enable(config: Mapping[str, Any] | None = None) -> CallableContextManager:
    """Stub"""


def enable(config=None):
    if config is None or isinstance(config, Mapping):
        return Enabled(config)
    # Treat as a decorator called on a function
    return Enabled()(config)


enabled = enable



from datetime import timedelta
from decimal import Decimal
from typing import Any, Dict


def get_schema():
    from django.apps import apps

    schema = {}
    for config in apps.get_app_configs():
        for model in config.get_models():
            schema[model._meta.db_table] = model_schema(model)
    return schema


def verbose_name(model):
    return str(model._meta.verbose_name).replace(" ", "_")


def model_schema(model):
    from django.contrib.contenttypes.fields import GenericForeignKey

    fields: Dict[str, Any] = {}
    schema = {
        "model_module": model.__module__,
        "model_name": model.__qualname__,
        "verbose_name": verbose_name(model),
        "fields": fields,
    }
    for field in model._meta.get_fields():
        try:
            column = field.column
        except AttributeError:
            # `ForeignObjectRel`
            continue

        if column is None:
            # `GenericRelation` or `GenericForeignKey`
            if isinstance(field, GenericForeignKey):
                fields[field.name] = gfk_schema(field)
            continue
        fields[column] = field_schema(field)

    return schema


def has_unique(field):
    if getattr(field, "_unique", False):
        return True

    meta = field.model._meta
    for fields in meta.unique_together:
        if field.name in fields:
            return True

    from django.db.models import UniqueConstraint

    for constraint in meta.constraints:
        if isinstance(constraint, UniqueConstraint) and field.name in constraint.fields:
            return True

    return False


def field_schema(field):
    from django.db.models.fields import NOT_PROVIDED

    field_class = type(field)
    schema = {
        "field_attname": field.attname,
        "field_name": field.name,
        "null": field.null,
        "primary_key": field.primary_key,
        "auto_created": field.auto_created,
        "is_relation": field.is_relation,
        "django_field": f"{field_class.__module__}.{field_class.__qualname__}",
        "has_unique": has_unique(field),
        "gfk_content_type_field": None,
        "gfk_object_id_field": None,
    }
    if field.default is not NOT_PROVIDED and not callable(field.default):
        schema["default"] = field_default(field.default)
    if hasattr(field, "auto_now"):
        schema["auto_now"] = field.auto_now
        schema["auto_now_add"] = field.auto_now_add
    if field.is_relation:
        schema["related_model"] = field.related_model._meta.db_table
        schema["related_pk"] = field.related_model._meta.pk.name

    return schema


def gfk_schema(field):
    return {
        "auto_created": field.auto_created,
        "django_field": "django.contrib.contenttypes.fields.GenericForeignKey",
        "field_attname": None,
        "field_name": field.name,
        "gfk_content_type_field": field.ct_field,
        "gfk_object_id_field": field.fk_field,
        "has_unique": has_unique(field),
        "is_relation": field.is_relation,
        "null": None,
        "primary_key": False,
        "related_model": None,
        "related_pk": None,
    }


def field_default(default):
    from django.db.models import Choices

    if isinstance(default, Choices):
        default = {
            "value": default._value_,
            "name": default._name_,
            "label": default.label,
            "choices_name": type(default).__qualname__,
            "choices_module": type(default).__module__,
        }
    elif isinstance(default, Decimal):
        default = {
            "value": str(default),
            "type": "Decimal",
        }
    elif isinstance(default, timedelta):
        default = {
            "value": {
                "days": default.days,
                "seconds": default.seconds,
                "microseconds": default.microseconds,
            },
            "type": "timedelta",
        }
    else:
        default = {"value": default}
    return default



from __future__ import annotations

import logging
import os
import platform
import types
from importlib import import_module
from typing import Any, Dict, List, Protocol, Tuple, Union


class FrameFilter(Protocol):
    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        pass  # pragma: no cover


ProtoFrameFilter = Union[str, FrameFilter, Dict[str, str]]


class FrameProcessor(Protocol):
    config: Dict[str, Any]
    co_names: Tuple[str, ...]

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        pass  # pragma: no cover

    def process(
        self,
        frame: types.FrameType,
        event: str,
        arg: object,
        call_frames: List[Tuple[types.FrameType, str]],
    ) -> Dict[str, Any]:
        pass  # pragma: no cover


logger = logging.getLogger("kolo")


class HasPath:
    def __init__(self, path: str):
        self.path = path

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        return self.path in frame.f_code.co_filename

    def __repr__(self):
        return f'HasPath("{self.path}")'

    def __eq__(self, other):
        return self.path == other.path


def build_frame_filter(filter: ProtoFrameFilter) -> FrameFilter:
    if isinstance(filter, str):
        return HasPath(filter)
    if isinstance(filter, dict):
        filter_path = filter["callable"]
        module_path, _sep, filter_name = filter_path.rpartition(".")
        module = import_module(module_path)
        return getattr(module, filter_name)
    return filter


def exec_filter(frame: types.FrameType, event: str, arg: object) -> bool:
    """
    Ignore a frame running a string executed using exec

    We can't show especially interesting information about it, so we skip it.

    A namedtuple is a common example of this case.
    """
    return frame.f_code.co_filename == "<string>"


def frozen_filter(frame: types.FrameType, event: str, arg: object) -> bool:
    """
    Ignore all frozen modules

    Frozen modules are compiled into the interpreter, so are almost
    always part of the standard library.

    To opt into showing a frozen module, specify it in include_frames.
    """
    return "<frozen " in frame.f_code.co_filename


class LibraryPathFilter:
    """
    Ignore library code

    We want to not show library calls, so attempt to filter them out here.
    """

    __slots__ = ("python_implementation", "system", "paths")

    paths: Tuple[str, ...]

    def __init__(self) -> None:
        # ‘CPython’, ‘IronPython’, ‘Jython’, ‘PyPy’.
        self.python_implementation = platform.python_implementation()

        if self.python_implementation not in ["CPython", "PyPy"]:
            logger.warning(
                "Kolo does not formally support this python_implementation: %s",
                self.python_implementation,
            )  # pragma: no cover

        # 'Linux', 'Darwin', 'Java', 'Windows'.
        # An empty string is returned if the value cannot be determined.
        self.system = platform.system()
        if self.system == "":
            logger.warning(
                "Kolo could not determine the current OS. If you're on Windows, capturing frames may be unreliable."
            )  # pragma: no cover

        # Cache all normalised path fragments.
        # Note: there is potentially a future where we can use `sys.base_prefix`,
        # `sys.exec_prefix`, `sys.prefix`, `sys.path`, and `sys.platlibdir` to amortise
        # some of this further based on the configured platform environment.
        default_paths = (
            os.path.normpath("lib/python"),
            os.path.normpath("/site-packages/"),
        )

        windows_paths = (os.path.normpath("/x64/lib/"),)

        pypy_paths = (
            os.path.normpath("lib/pypy"),
            os.path.normpath("versions/pypy"),
            os.path.normpath("/PyPy/"),
        )

        if self.python_implementation == "PyPy" and self.system == "Windows":
            self.paths = default_paths + windows_paths + pypy_paths
        elif self.system == "Windows":
            self.paths = default_paths + windows_paths
        elif self.python_implementation == "PyPy":
            self.paths = default_paths + pypy_paths
        else:
            self.paths = default_paths

    def __call__(self, frame: types.FrameType, *args, **kwargs) -> bool:
        filepath = frame.f_code.co_filename
        # Avoid any() + a <genexp> and break ASAP, it's marginally faster.
        for path in self.paths:
            if path in filepath:
                return True

        # Something like \python\python310\lib\ on windows
        if self.system == "Windows":
            return (
                "\\python\\" in filepath or "\\Python\\" in filepath or "\\pypy\\"
            ) and ("\\lib\\" in filepath or "\\Lib\\" in filepath)
        else:
            return False


library_filter = LibraryPathFilter().__call__



from __future__ import annotations

import os
import threading
import time
from typing import Dict

import ulid


class PytestFilter:
    event_types = {
        "pytest_runtest_logstart": "start_test",
        "pytest_runtest_logfinish": "end_test",
    }
    co_names = tuple(event_types)
    pytest_plugin_filename = os.path.normpath("kolo/pytest_plugin.py")

    def __init__(self, config) -> None:
        self.config = config
        self._frame_ids: Dict[int, str] = {}

    def __call__(self, frame, event, arg):  # pragma: no cover
        return (
            event == "call"
            and self.pytest_plugin_filename in frame.f_code.co_filename
            and frame.f_code.co_name in self.event_types
        )

    def process(self, frame, event, arg, call_frames):  # pragma: no cover
        timestamp = time.time()
        thread = threading.current_thread()
        native_id = getattr(thread, "native_id", None)
        co_name = frame.f_code.co_name
        location = frame.f_locals["location"]
        if co_name == "pytest_runtest_logstart":
            frame_id = f"frm_{ulid.new()}"
            self._frame_ids[id(location)] = frame_id
        else:
            frame_id = self._frame_ids[id(location)]

        filename, lineno, test = location
        test_class, _sep, test_name = test.rpartition(".")
        return {
            "frame_id": frame_id,
            "thread": thread.name,
            "thread_native_id": native_id,
            "type": self.event_types[co_name],
            "test_name": test_name,
            "test_class": test_class if test_class else None,
            "timestamp": timestamp,
        }



from __future__ import annotations

import threading
import time
import types
from typing import Any, Dict, List, Literal, Optional, Tuple, TypedDict

import ulid


class CeleryJob(TypedDict, total=False):
    frame_id: str
    name: str
    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]
    thread: str
    thread_native_id: Optional[int]
    timestamp: float
    type: Literal["background_job", "background_job_end"]
    subtype: Literal["celery"]


class CeleryFilter:
    co_names: Tuple[str, ...] = ("apply_async",)

    def __init__(self, config) -> None:
        self.config = config
        self._frame_ids: Dict[int, str] = {}

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        filepath = frame.f_code.co_filename
        return (
            "celery" in filepath
            and "sentry_sdk" not in filepath
            and "apply_async" in frame.f_code.co_name
        )

    def process(
        self,
        frame: types.FrameType,
        event: str,
        arg: object,
        call_frames: List[Tuple[types.FrameType, str]],
    ):
        timestamp = time.time()
        celery_job: CeleryJob
        frame_locals = frame.f_locals
        task_name = frame_locals["self"].name
        task_args = frame_locals["args"]
        task_kwargs = frame_locals["kwargs"]
        thread = threading.current_thread()

        if event == "call":
            frame_id = f"frm_{ulid.new()}"
            self._frame_ids[id(frame)] = frame_id
            celery_job = {
                "frame_id": frame_id,
                "name": task_name,
                "args": task_args,
                "kwargs": task_kwargs,
                "thread": thread.name,
                "thread_native_id": thread.native_id,
                "timestamp": timestamp,
                "type": "background_job",
                "subtype": "celery",
            }
            return celery_job

        assert event == "return"

        celery_job = {
            "frame_id": self._frame_ids[id(frame)],
            "name": task_name,
            "args": task_args,
            "kwargs": task_kwargs,
            "thread": thread.name,
            "thread_native_id": thread.native_id,
            "timestamp": timestamp,
            "type": "background_job_end",
            "subtype": "celery",
        }
        return celery_job



from __future__ import annotations

import os
import types

KOLO_PATHS = (
    "/kolo/config.py",
    "/kolo/db.py",
    "/kolo/django_schema.py",
    "/kolo/filters/",
    "/kolo/generate_tests/",
    "/kolo/git.py",
    "/kolo/__init__.py",
    "/kolo/__main__.py",
    "/kolo/middleware.py",
    "/kolo/profiler.py",
    "/kolo/pytest_plugin.py",
    "/kolo/serialize.py",
    "/kolo/utils.py",
    "/kolo/version.py",
)
KOLO_PATHS = tuple(os.path.normpath(path) for path in KOLO_PATHS)


def kolo_filter(frame: types.FrameType, event: str, arg: object) -> bool:
    """Don't profile kolo code"""
    filename = frame.f_code.co_filename
    return any(path in filename for path in KOLO_PATHS)



from __future__ import annotations

import gzip
import logging
import os
import types
from datetime import date, datetime, time
from decimal import Decimal
from email.message import Message
from email.utils import collapse_rfc2231_value
from typing import TYPE_CHECKING, Any, Dict, TypedDict, TypeVar

try:
    from django.db.models import QuerySet
except ImportError:
    QuerySet = None  # type: ignore

import msgpack
import simplejson

logger = logging.getLogger("kolo")


if TYPE_CHECKING:
    from django.http import HttpRequest, HttpResponse, StreamingHttpResponse


class UserCodeCallSite(TypedDict):
    line_number: int
    call_frame_id: str


Local = TypeVar("Local")


SERIALIZE_PATH = os.path.normpath("kolo/serialize.py")


# TODO: Make these threadlocals when we support multithreading
QUERYSET_PATCHED = False
IN_KOLO_PROFILER = False


def monkeypatch_queryset_repr():
    global QUERYSET_PATCHED
    if QUERYSET_PATCHED:
        return

    try:
        from django.db.models import QuerySet
    except ImportError:  # pragma: no cover
        QUERYSET_PATCHED = True
        return

    old_repr = QuerySet.__repr__

    def new_repr(queryset):
        if queryset._result_cache is None and IN_KOLO_PROFILER:
            return f"Unevaluated queryset for: {queryset.model}"
        return old_repr(queryset)

    QuerySet.__repr__ = new_repr  # type: ignore
    QUERYSET_PATCHED = True


def json_default(obj):
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    try:
        return repr(obj)
    except Exception:
        return "SerializationError"


def dump_json(data):
    global IN_KOLO_PROFILER
    IN_KOLO_PROFILER = True
    try:
        return simplejson.dumps(
            data,
            skipkeys=True,
            ignore_nan=True,
            namedtuple_as_object=False,
            default=json_default,
        )
    finally:
        IN_KOLO_PROFILER = False


def msgpack_encode_hook(obj):
    if isinstance(obj, tuple):
        return list(obj)

    if isinstance(obj, dict):
        return dict(obj)

    if isinstance(obj, datetime):
        return msgpack.ExtType(1, obj.isoformat().encode("utf-8"))

    if isinstance(obj, date):
        return msgpack.ExtType(2, obj.isoformat().encode("utf-8"))

    if isinstance(obj, time):
        return msgpack.ExtType(3, obj.isoformat().encode("utf-8"))

    if isinstance(obj, int):
        num_bytes = (obj.bit_length() + 7) // 8 + 1
        return msgpack.ExtType(4, obj.to_bytes(num_bytes, "big", signed=True))

    if isinstance(obj, Decimal):
        return msgpack.ExtType(5, str(obj).encode("utf-8"))

    if (
        QuerySet is not None
        and isinstance(obj, QuerySet)  # type: ignore[misc]
        and obj._result_cache is None
    ):
        data = f"Unevaluated queryset for: {obj.model}".encode("utf-8")
        return msgpack.ExtType(0, data)

    try:
        return msgpack.ExtType(0, repr(obj).encode("utf-8"))
    except Exception:
        path = f"{obj.__module__}.{obj.__class__.__qualname__}"
        return msgpack.ExtType(127, path.encode("utf-8"))


def msgpack_decode_hook(code, data):
    if code == 0:
        return data.decode("utf-8")
    if code == 1:
        return datetime.fromisoformat(data.decode("utf-8"))
    if code == 2:
        return date.fromisoformat(data.decode("utf-8"))
    if code == 3:
        return time.fromisoformat(data.decode("utf-8"))
    if code == 4:
        return int.from_bytes(data, "big", signed=True)
    if code == 5:
        return Decimal(data.decode("utf-8"))
    if code == 127:
        data = data.decode("utf-8")
        return f"SerializationError: {data}"
    raise ValueError(f"Unknown msgpack extension. Code: {code}")


def dump_msgpack(data):
    global IN_KOLO_PROFILER
    IN_KOLO_PROFILER = True
    try:
        return msgpack.packb(data, default=msgpack_encode_hook, strict_types=True)
    finally:
        IN_KOLO_PROFILER = False


def dump_data(data, config):
    if config.get("use_msgpack", True):
        return dump_msgpack(data)
    return dump_json(data)


def load_msgpack(data):
    return msgpack.unpackb(data, strict_map_key=False, ext_hook=msgpack_decode_hook)


def decode_header_value(bytes_or_str: bytes | str) -> str:
    """
    Convert a bytes header value to text.

    Valid header values are expected to be ascii in modern times, but
    ISO-8859-1 (latin1) has historically been allowed.

    https://datatracker.ietf.org/doc/html/rfc7230#section-3.2.4
    """
    if isinstance(bytes_or_str, bytes):
        return bytes_or_str.decode("latin1")
    return bytes_or_str


def frame_path(frame: types.FrameType) -> str:
    path = frame.f_code.co_filename
    try:
        relative_path = os.path.relpath(path)
    except ValueError:
        relative_path = path
    return f"{relative_path}:{frame.f_lineno}"


def decode_body(body: Any, request_headers: Dict[str, str]) -> Any:
    """Convert a request body into a json-serializable form."""
    if isinstance(body, bytes):
        content_type = request_headers.get("Content-Type", "")
        m = Message()
        m["content-type"] = content_type
        charset = collapse_rfc2231_value(m.get_param("charset", "utf-8"))
        try:
            return body.decode(charset)
        except UnicodeDecodeError:
            return "<Binary request body>"
    return body


def get_content(response: HttpResponse | StreamingHttpResponse) -> str:
    if response.streaming:
        return "<Streaming Response>"

    if TYPE_CHECKING:
        assert isinstance(response, HttpResponse)
    content_encoding = response.get("Content-Encoding")
    if content_encoding == "gzip":
        content = gzip.decompress(response.content)
    else:
        content = response.content
    try:
        return content.decode(response.charset)
    except UnicodeDecodeError:
        return f"<Response with invalid charset ({response.charset})>"


def get_request_body(request: "HttpRequest") -> str:
    from django.http.request import RawPostDataException

    try:
        return request.body.decode("utf-8")
    except UnicodeDecodeError:  # pragma: no cover
        return "<Binary request body>"
    except RawPostDataException:
        return "<Request data already read>"



from __future__ import annotations

import os
import threading
import time
import types
from contextlib import suppress
from typing import Dict, List, Tuple

import ulid

from ..serialize import get_content, get_request_body


class DjangoFilter:
    co_names: Tuple[str, ...] = ("get_response",)
    kolo_middleware_filename = os.path.normpath("/kolo/middleware.py")

    def __init__(self, config) -> None:
        self.config = config
        self.timestamp: float
        self._frame_ids: Dict[int, str] = {}

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        return (
            self.kolo_middleware_filename in frame.f_code.co_filename
            and frame.f_code.co_name in self.co_names
        )

    def process(
        self,
        frame: types.FrameType,
        event: str,
        arg: object,
        call_frames: List[Tuple[types.FrameType, str]],
    ):
        thread = threading.current_thread()
        if event == "call":
            frame_id = f"frm_{ulid.new()}"
            self._frame_ids[id(frame)] = frame_id
            self.timestamp = time.time()
            request = frame.f_locals["request"]
            return {
                "body": get_request_body(request),
                "frame_id": frame_id,
                "headers": dict(request.headers),
                "method": request.method,
                "path_info": request.path_info,
                # dict(request.POST) can give confusing results due to MultivalueDict
                "post_data": dict(request.POST),
                "query_params": dict(request.GET),
                "scheme": request.scheme,
                "thread": thread.name,
                "thread_native_id": thread.native_id,
                "timestamp": self.timestamp,
                "type": "django_request",
            }
        elif event == "return":  # pragma: no branch
            timestamp = time.time()
            duration = timestamp - self.timestamp
            ms_duration = round(duration * 1000, 2)

            request = frame.f_locals["request"]
            match = request.resolver_match
            if match:  # match is None if this is a 404
                url_pattern = {
                    "namespace": match.namespace,
                    "route": match.route,
                    "url_name": match.url_name,
                    "view_qualname": match._func_path,
                }
            else:
                url_pattern = None

            response = frame.f_locals["response"]
            return {
                "frame_id": self._frame_ids[id(frame)],
                "ms_duration": ms_duration,
                "status_code": response.status_code,
                "content": get_content(response),
                "headers": dict(response.items()),
                "thread": thread.name,
                "thread_native_id": thread.native_id,
                "timestamp": timestamp,
                "type": "django_response",
                "url_pattern": url_pattern,
            }


class DjangoTemplateFilter:
    co_names: Tuple[str, ...] = ("render",)
    django_template_filename = os.path.normpath("django/template/backends/django.py")

    def __init__(self, config) -> None:
        self.config = config
        self.timestamp: float
        self._frame_ids: Dict[int, str] = {}

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        return (
            self.django_template_filename in frame.f_code.co_filename
            and frame.f_code.co_name in self.co_names
        )

    def process(
        self,
        frame: types.FrameType,
        event: str,
        arg: object,
        call_frames: List[Tuple[types.FrameType, str]],
    ):
        timestamp = time.time()
        thread = threading.current_thread()
        context = frame.f_locals["context"]
        with suppress(AttributeError):
            context = context.flatten()
        if event == "call":
            frame_id = f"frm_{ulid.new()}"
            self._frame_ids[id(frame)] = frame_id
            _type = "django_template_start"
        else:
            frame_id = self._frame_ids[id(frame)]
            _type = "django_template_end"

        return {
            "context": context,
            "frame_id": frame_id,
            "template": frame.f_locals["self"].template.name,
            "thread": thread.name,
            "timestamp": timestamp,
            "type": _type,
        }


class DjangoSetupFilter:
    co_names: Tuple[str, ...] = ("setup",)
    django_init_filename = os.path.normpath("django/__init__.py")

    def __init__(self, config) -> None:
        self.config = config
        self.timestamp: float
        self._frame_ids: Dict[int, str] = {}

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        return (
            self.django_init_filename in frame.f_code.co_filename
            and frame.f_code.co_name in self.co_names
        )

    def process(
        self,
        frame: types.FrameType,
        event: str,
        arg: object,
        call_frames: List[Tuple[types.FrameType, str]],
    ):
        timestamp = time.time()
        thread = threading.current_thread()
        if event == "call":
            frame_id = f"frm_{ulid.new()}"
            self._frame_ids[id(frame)] = frame_id
            _type = "django_setup_start"
        else:
            frame_id = self._frame_ids[id(frame)]
            _type = "django_setup_end"

        return {
            "frame_id": frame_id,
            "thread": thread.name,
            "timestamp": timestamp,
            "type": _type,
        }


class DjangoChecksFilter:
    co_names: Tuple[str, ...] = ("run_checks",)
    django_checks_filename = os.path.normpath("django/core/checks/registry.py")

    def __init__(self, config) -> None:
        self.config = config
        self.timestamp: float
        self._frame_ids: Dict[int, str] = {}

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        return (
            self.django_checks_filename in frame.f_code.co_filename
            and frame.f_code.co_name in self.co_names
        )

    def process(
        self,
        frame: types.FrameType,
        event: str,
        arg: object,
        call_frames: List[Tuple[types.FrameType, str]],
    ):
        timestamp = time.time()
        thread = threading.current_thread()
        if event == "call":
            frame_id = f"frm_{ulid.new()}"
            self._frame_ids[id(frame)] = frame_id
            _type = "django_checks_start"
        else:
            frame_id = self._frame_ids[id(frame)]
            _type = "django_checks_end"

        return {
            "frame_id": frame_id,
            "thread": thread.name,
            "timestamp": timestamp,
            "type": _type,
        }


class DjangoTestDBFilter:
    co_names: Tuple[str, ...] = ("create_test_db",)
    django_checks_filename = os.path.normpath("django/db/backends/base/creation.py")

    def __init__(self, config) -> None:
        self.config = config
        self.timestamp: float
        self._frame_ids: Dict[int, str] = {}

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        return (
            self.django_checks_filename in frame.f_code.co_filename
            and frame.f_code.co_name in self.co_names
        )

    def process(
        self,
        frame: types.FrameType,
        event: str,
        arg: object,
        call_frames: List[Tuple[types.FrameType, str]],
    ):
        timestamp = time.time()
        thread = threading.current_thread()
        if event == "call":
            frame_id = f"frm_{ulid.new()}"
            self._frame_ids[id(frame)] = frame_id
            _type = "django_create_test_db_start"
        else:
            frame_id = self._frame_ids[id(frame)]
            _type = "django_create_test_db_end"

        return {
            "frame_id": frame_id,
            "thread": thread.name,
            "timestamp": timestamp,
            "type": _type,
        }



from importlib import import_module
from typing import List

from kolo.generate_tests.queries import DjangoField


class Factory:
    def __init__(self, factory: str, values: List[DjangoField], variable_name: str):
        self.factory = factory
        self.values = values
        self.variable_name = variable_name

    @property
    def template_parts(self):
        yield f"{self.variable_name} = {self.factory}.create("
        for field in self.values:
            yield f"{field.name}={field.value_repr},"
        yield ")"


def import_factory(factory_path):
    try:
        module_path, factory_name = factory_path.rsplit(".", 1)
    except ValueError:
        raise ImportError(f"Could not import `{factory_path}`.")

    module = import_module(module_path)
    factory = getattr(module, factory_name)
    return factory, f"from {module_path} import {factory_name}"


def import_factories(config):
    factory_configs = config.get("test_generation", {}).get("factories", {})
    factories = {}
    for factory_config in factory_configs:
        factory_path = factory_config["path"]
        factory, import_path = import_factory(factory_path)
        factory_config["factory"] = factory
        factory_config["name"] = factory.__name__
        factory_config["import_path"] = import_path
        model = factory._meta.model
        model_path = f"{model.__name__}"
        factories[model_path] = factory_config
    return factories


def filter_defaults(fields, factory_class):
    filtered = []
    for field in fields:
        name = field.name
        try:
            value = getattr(factory_class, name)
        except AttributeError:
            filtered.append(field)
            continue

        if field.value_repr == repr(value):
            continue
        filtered.append(field)
    return filtered


def build_factory(model, config):
    if config.get("pk", False):
        values = [model.primary_key]
        values.extend(model.values)
    else:
        values = model.values

    factory_class = config["factory"]
    fields = filter_defaults(values, factory_class)

    factory = Factory(
        factory=config["name"],
        values=fields,
        variable_name=model.variable_name,
    )
    return factory


def build_factories(fixtures, configs):
    factories = []
    imports = set()
    for create_model in fixtures:
        model = create_model.model
        try:
            config = configs[model]
        except KeyError:
            factories.append(create_model)
            continue

        factory = build_factory(create_model, config)
        factories.append(factory)
        imports.add(config["import_path"])
    return factories, imports



from __future__ import annotations

import logging
import os
import threading
import time
import types
from typing import List, Tuple

import ulid

formatter = logging.Formatter()


class LoggingFilter:
    co_names: Tuple[str, ...] = ("_log",)
    logging_filename = os.path.normpath("/logging/")

    def __init__(self, config) -> None:
        self.config = config

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        return (
            event == "return"
            and self.logging_filename in frame.f_code.co_filename
            and frame.f_code.co_name in self.co_names
        )

    def process(
        self,
        frame: types.FrameType,
        event: str,
        arg: object,
        call_frames: List[Tuple[types.FrameType, str]],
    ):
        timestamp = time.time()
        thread = threading.current_thread()
        native_id = getattr(thread, "native_id", None)
        frame_locals = frame.f_locals
        exc_info = frame_locals["exc_info"]
        traceback = None if exc_info is None else formatter.formatException(exc_info)
        extra = frame_locals["extra"]
        if call_frames:
            call_frame, call_frame_id = call_frames[-1]
            user_code_call_site = {
                "call_frame_id": call_frame_id,
                "line_number": call_frame.f_lineno,
            }
        else:
            user_code_call_site = None
        return {
            "args": frame_locals["args"],
            "extra": extra,
            "frame_id": f"frm_{ulid.new()}",
            "level": logging.getLevelName(frame_locals["level"]),
            "msg": frame_locals["msg"],
            "stack": formatter.formatStack(frame_locals["sinfo"]),
            "thread": thread.name,
            "thread_native_id": thread.native_id,
            "timestamp": timestamp,
            "traceback": traceback,
            "type": "log_message",
            "user_code_call_site": user_code_call_site,
        }



from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import Any, Mapping, MutableMapping

try:
    import tomllib
except ImportError:
    import tomli as tomllib

import toolz
from cerberus import Validator

logger = logging.getLogger("kolo")


class KoloWriteError(Exception):
    pass


class InvalidConfig(Exception):
    pass


schema = {
    "filters": {
        "type": "dict",
        "schema": {
            "include_frames": {"type": "list"},
            "ignore_frames": {"type": "list"},
            "ignore_request_paths": {"type": "list"},
        },
    },
    "sqlite_busy_timeout": {"type": "integer"},
    "test_generation": {
        "type": "dict",
        "schema": {
            "factories": {"type": "list"},
            "trace_processors": {"type": "list"},
        },
    },
    "threading": {"type": "boolean"},
    "use_msgpack": {"type": "boolean"},
    "use_rust": {"type": "boolean"},
    "web_experience": {"type": "boolean"},
}
pyproject_schema = {
    "tool": {
        "type": "dict",
        "schema": {
            "kolo": {
                "type": "dict",
                "schema": schema,
                "allow_unknown": False,
            },
        },
    },
}
validator = Validator(schema, allow_unknown=False)
pyproject_validator = Validator(pyproject_schema, allow_unknown=True)


def clear_errors(config: MutableMapping[str, Any], errors) -> None:
    for error in errors:
        key = error.document_path[-1]
        if error.info:
            for info in error.info:
                clear_errors(config[key], info)
        else:
            del config[key]


def load_config_from_toml(path: Path) -> MutableMapping[str, Any]:
    try:
        with open(path, "rb") as conf:
            config = tomllib.load(conf)
    except FileNotFoundError:
        logger.info(
            'Kolo config file "%s" not found. Using default configuration.', path
        )
        return {}
    if not validator.validate(config):
        logger.warning("Kolo config file has errors: %s", validator.errors)
        clear_errors(config, validator._errors)
    return config


def load_config_from_pyproject_toml(path: Path) -> MutableMapping[str, Any]:
    try:
        with open(path, "rb") as conf:
            config = tomllib.load(conf)
    except FileNotFoundError:
        logger.info("%s config file not found. Trying next configuration option.", path)
        raise InvalidConfig()
    if not pyproject_validator.validate(config):
        logger.warning(
            "pyproject.toml config file has errors: %s", pyproject_validator.errors
        )
        clear_errors(config, pyproject_validator._errors)
    try:
        return config["tool"]["kolo"]
    except KeyError:
        raise InvalidConfig()


def create_kolo_directory() -> Path:
    """
    Create the kolo directory and contents if they do not exist

    Returns the path to the .kolo directory for convenience.
    """
    kolo_directory = (Path(os.environ.get("KOLO_PATH", ".")) / ".kolo").resolve()
    try:
        kolo_directory.mkdir(parents=True, exist_ok=True)
    except Exception as e:
        message = f"Could not create .kolo directory at {kolo_directory}."
        raise KoloWriteError(message) from e

    gitignore_path = kolo_directory / ".gitignore"
    try:
        with open(gitignore_path, "w") as gitignore:
            gitignore.write("db.sqlite3\n")
            gitignore.write("db.sqlite3-shm\n")
            gitignore.write("db.sqlite3-wal\n")
            gitignore.write(".gitignore\n")
    except Exception as e:
        message = f"Could not write to {gitignore_path}."
        raise KoloWriteError(message) from e
    return kolo_directory.resolve()


def merge_or_last(args):
    if all(isinstance(arg, dict) for arg in args):
        return toolz.merge(*args)
    return args[-1]


def load_config(
    extra_config: Mapping[str, Any] | None = None
) -> MutableMapping[str, Any]:
    kolo_directory = create_kolo_directory()
    pyproject_toml_path = Path("pyproject.toml")
    try:
        config = load_config_from_pyproject_toml(pyproject_toml_path)
    except InvalidConfig:
        config = load_config_from_toml(kolo_directory / "config.toml")
    if extra_config:
        return toolz.merge_with(merge_or_last, config, extra_config)
    return config



import os
import threading
import time
import types
from typing import Dict, List, Literal, Set, Tuple

import ulid

from .types import BaseOutboundRequest, BaseOutboundResponse


class OutboundRequest(BaseOutboundRequest):
    subtype: Literal["httpx"]


class OutboundResponse(BaseOutboundResponse):
    subtype: Literal["httpx"]


class HttpxFilter:
    co_names: Tuple[str, ...] = ("send",)
    httpx_filename = os.path.normpath("httpx/_client.py")

    def __init__(self, config) -> None:
        self.config = config
        self._frame_ids: Dict[int, str] = {}
        self._request_ids: Set[int] = set()

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        return self.httpx_filename in frame.f_code.co_filename

    def process(
        self,
        frame: types.FrameType,
        event: str,
        arg: object,
        call_frames: List[Tuple[types.FrameType, str]],
    ):
        timestamp = time.time()
        thread = threading.current_thread()
        frame_locals = frame.f_locals
        # Track `request_id` to strip out unwanted resume frames in the async case.
        # Unfortunately `request_id` isn't guaranteed to be unique if the user chooses
        # to re-use a `request` instance, but I haven't thought of a way to do better.
        # We can't use `self` (`Client` or `AsyncClient` because that's even more likely
        # to be re-used.
        request_id = id(frame_locals["request"])
        request = frame_locals["request"]
        method = request.method
        url = str(request.url)

        if event == "call":
            if request_id in self._request_ids:
                # If the `request_id` is re-used, assume we're resuming the coroutine
                return
            self._request_ids.add(request_id)
            frame_id = f"frm_{ulid.new()}"
            self._frame_ids[id(frame)] = frame_id
            api_request: OutboundRequest = {
                "body": None,
                "frame_id": frame_id,
                "headers": dict(request.headers),
                "method": method,
                "method_and_full_url": f"{method} {url}",
                "subtype": "httpx",
                "thread": thread.name,
                "thread_native_id": thread.native_id,
                "timestamp": timestamp,
                "type": "outbound_http_request",
                "url": url,
            }
            return api_request

        assert event == "return"

        api_response: OutboundResponse

        if arg is None:
            # Probably an exception
            api_response = {
                "type": "outbound_http_response",
                "body": None,
                "frame_id": self._frame_ids[id(frame)],
                "headers": None,
                "method": method,
                "method_and_full_url": f"{method} {url}",
                "status_code": None,
                "subtype": "httpx",
                "thread": thread.name,
                "thread_native_id": thread.native_id,
                "timestamp": timestamp,
                "url": url,
            }
            return api_response

        from httpx._models import Response

        if not isinstance(arg, Response):
            # If we don't have a `Response`, assume we're suspending the coroutine
            return

        self._request_ids.discard(request_id)

        from httpx._exceptions import ResponseNotRead

        try:
            body = arg.text
        except ResponseNotRead:  # pragma: no cover
            body = None

        api_response = {
            "type": "outbound_http_response",
            "body": body,
            "frame_id": self._frame_ids[id(frame)],
            "headers": dict(arg.headers),
            "method": method,
            "method_and_full_url": f"{method} {url}",
            "status_code": arg.status_code,
            "subtype": "httpx",
            "thread": thread.name,
            "thread_native_id": thread.native_id,
            "timestamp": timestamp,
            "url": url,
        }
        return api_response



from __future__ import annotations


def pytest_runtest_logstart(nodeid, location):
    """Stub start test hook for PytestFilter"""


def pytest_runtest_logfinish(nodeid, location):
    """Stub end test hook for PytestFilter"""



from __future__ import annotations

import os
import threading
import time
import types
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
    TypedDict,
)

import ulid

from ..serialize import UserCodeCallSite

if TYPE_CHECKING:
    from django.db.models.sql.compiler import SQLUpdateCompiler


class QueryStart(TypedDict):
    database: str
    frame_id: str
    user_code_call_site: UserCodeCallSite | None
    call_timestamp: float
    thread: str
    thread_native_id: Optional[int]
    timestamp: float
    type: Literal["start_sql_query"]


class QueryEnd(TypedDict, total=False):
    database: str
    frame_id: str
    query: str | None
    query_data: List[Any] | str | None
    query_template: str | None
    return_timestamp: float
    thread: str
    thread_native_id: Optional[int]
    timestamp: float
    type: Literal["end_sql_query"]


class SQLQueryFilter:
    co_names: Tuple[str, ...] = ("execute_sql",)
    django_sql_filename = os.path.normpath("/django/db/models/sql/compiler.py")
    klass: Type[SQLUpdateCompiler] | None = None

    def __init__(self, config) -> None:
        self.config = config
        self._frame_ids: Dict[int, str] = {}

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        if (
            self.django_sql_filename in frame.f_code.co_filename
            and frame.f_code.co_name in self.co_names
        ):  # pragma: no cover
            # Doing the import once and binding is substantially faster than going
            # through the import system every time.
            # See:
            # https://github.com/django/asgiref/issues/269
            # https://github.com/django/asgiref/pull/288
            # A more battle-tested (slightly slower/more complex) alternative would be:
            # https://github.com/django/django/pull/14850
            # https://github.com/django/django/pull/14858
            # https://github.com/django/django/pull/14931
            if SQLQueryFilter.klass is None:
                from django.db.models.sql.compiler import SQLUpdateCompiler

                SQLQueryFilter.klass = SQLUpdateCompiler
            return frame.f_code is not SQLQueryFilter.klass.execute_sql.__code__
        return False

    def process(
        self,
        frame: types.FrameType,
        event: str,
        arg: Any,
        call_frames: List[Tuple[types.FrameType, str]],
    ):
        timestamp = time.time()
        thread = threading.current_thread()
        database = frame.f_locals["self"].connection.vendor
        if event == "call":
            frame_id = f"frm_{ulid.new()}"
            self._frame_ids[id(frame)] = frame_id
            if call_frames:
                call_frame, call_frame_id = call_frames[-1]
                user_code_call_site: UserCodeCallSite | None = {
                    "call_frame_id": call_frame_id,
                    "line_number": call_frame.f_lineno,
                }
            else:
                user_code_call_site = None

            query_start: QueryStart = {
                "database": database,
                "frame_id": frame_id,
                "user_code_call_site": user_code_call_site,
                "call_timestamp": time.time(),
                "thread": thread.name,
                "thread_native_id": thread.native_id,
                "timestamp": timestamp,
                "type": "start_sql_query",
            }
            return query_start

        assert event == "return"

        try:
            sql = frame.f_locals["sql"]
            params = frame.f_locals["params"]
        except KeyError:
            query_template = None
            query = None
        else:
            if sql == "":
                query_template = None
                query = None
            else:
                cursor = frame.f_locals["cursor"]
                ops = frame.f_locals["self"].connection.ops
                query_template = sql.strip()
                query = ops.last_executed_query(cursor, sql, params).strip()

        query_end: QueryEnd = {
            "database": database,
            "frame_id": self._frame_ids[id(frame)],
            "return_timestamp": timestamp,
            "query_template": query_template,
            "query": query,
            "query_data": arg,
            "thread": thread.name,
            "thread_native_id": thread.native_id,
            "timestamp": timestamp,
            "type": "end_sql_query",
        }
        return query_end



from typing import Dict, Literal, Optional, TypedDict


class BaseOutboundRequest(TypedDict):
    body: Optional[str]
    frame_id: str
    headers: Dict[str, str]
    method: str
    method_and_full_url: str
    thread: str
    thread_native_id: Optional[int]
    timestamp: float
    type: Literal["outbound_http_request"]
    url: str


class BaseOutboundResponse(TypedDict):
    body: Optional[str]
    frame_id: str
    headers: Optional[Dict[str, str]]
    method: str
    method_and_full_url: str
    status_code: Optional[int]
    thread: str
    thread_native_id: Optional[int]
    timestamp: float
    type: Literal["outbound_http_response"]
    url: str



from __future__ import annotations

import os
import threading
import time
from typing import Dict

import ulid


class UnitTestFilter:
    event_types = {
        "startTest": "start_test",
        "stopTest": "end_test",
    }
    co_names = tuple(event_types)
    unittest_result_filename = os.path.normpath("unittest/result.py")

    def __init__(self, config) -> None:
        self.config = config
        self._frame_ids: Dict[int, str] = {}

    def __call__(self, frame, event, arg):  # pragma: no cover
        return (
            event == "call"
            and self.unittest_result_filename in frame.f_code.co_filename
            and frame.f_code.co_name in self.event_types
        )

    def process(self, frame, event, arg, call_frames):  # pragma: no cover
        timestamp = time.time()
        thread = threading.current_thread()
        testcase = frame.f_locals["test"]
        co_name = frame.f_code.co_name
        if co_name == "startTest":
            frame_id = f"frm_{ulid.new()}"
            self._frame_ids[id(testcase)] = frame_id
        else:
            frame_id = self._frame_ids[id(testcase)]
        return {
            "frame_id": frame_id,
            "thread": thread.name,
            "thread_native_id": thread.native_id,
            "type": self.event_types[co_name],
            "test_name": testcase._testMethodName,
            "test_class": testcase.__class__.__qualname__,
            "timestamp": timestamp,
        }



from __future__ import annotations

import json
import logging
import sqlite3
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Tuple

from .config import create_kolo_directory
from .serialize import load_msgpack

logger = logging.getLogger("kolo")


class SchemaNotFoundError(Exception):
    pass


class TraceNotFoundError(Exception):
    pass


@contextmanager
def db_connection(db_path, timeout=60):
    """
    Wrap sqlite's connection for use as a context manager

    Commits all changes if no exception is raised.
    Always closes the connection after the context manager exits.
    """
    connection = sqlite3.connect(str(db_path), isolation_level=None, timeout=timeout)
    try:
        connection.execute("pragma journal_mode=wal")
    finally:
        connection.close()

    connection = sqlite3.connect(str(db_path), isolation_level=None, timeout=timeout)
    try:
        with connection:
            yield connection
    finally:
        connection.close()


def get_db_path() -> Path:
    return create_kolo_directory() / "db.sqlite3"


def create_traces_table(connection) -> None:
    create_table_query = """
    CREATE TABLE IF NOT EXISTS traces (
        id text PRIMARY KEY NOT NULL,
        created_at TEXT DEFAULT (STRFTIME('%Y-%m-%d %H:%M:%f', 'NOW')) NOT NULL,
        data text NULL,
        msgpack blob NULL
    );
    """
    create_timestamp_index_query = """
        CREATE INDEX IF NOT EXISTS
        idx_traces_created_at
        ON traces (created_at);
        """
    invocations_exists_query = """
    SELECT EXISTS(
        SELECT 1 FROM sqlite_master
        WHERE type="table" AND name="invocations"
    );
    """
    copy_invocations_query = """
    INSERT OR IGNORE INTO traces (id, created_at, data, msgpack)
        SELECT id, created_at, data, NULL FROM invocations;
    """
    drop_invocations_query = "DROP TABLE invocations;"

    connection.execute(create_table_query)
    connection.execute(create_timestamp_index_query)
    cursor = connection.execute(invocations_exists_query)
    if cursor.fetchone()[0]:
        connection.execute(copy_invocations_query)
        connection.execute(drop_invocations_query)


def setup_db() -> Path:
    db_path = get_db_path()

    with db_connection(db_path) as connection:
        create_traces_table(connection)

    return db_path


def save_trace_in_sqlite(
    db_path: Path,
    trace_id: str,
    json_string: str | None = None,
    msgpack: bytes | None = None,
    *,
    ignore_errors: bool = True,
    created_at: datetime | None = None,
    timeout=60,
) -> None:
    ignore = " OR IGNORE" if ignore_errors else ""
    _columns = ["id", "data", "msgpack"]
    values: list[object] = [trace_id, json_string, msgpack]
    if created_at is not None:
        _columns.append("created_at")
        values.append(created_at)
    columns = ", ".join(_columns)
    params = ",".join(["?" for _ in _columns])

    insert_sql = f"INSERT{ignore} INTO traces({columns}) VALUES({params})"

    # We can't reuse a connection
    # because we're in a new thread
    with db_connection(db_path, timeout) as connection:
        try:
            connection.execute(insert_sql, values)
        except (sqlite3.DataError, sqlite3.InterfaceError):
            # DataError on python 3.11+
            # InterfaceError on python 3.10-
            logger.exception("The generated trace was too big to store in sqlite.")


def load_trace(msgpack_data: bytes | None, json_data: str | None):
    if msgpack_data is None:
        assert json_data is not None
        return json.loads(json_data)
    return load_msgpack(msgpack_data)


def load_trace_from_db(db_path: Path, trace_id: str) -> Tuple[bytes | None, str | None]:
    fetch_sql = "SELECT msgpack, data FROM traces WHERE id = ?"

    with db_connection(db_path) as connection:
        cursor = connection.execute(fetch_sql, (trace_id,))
        row = cursor.fetchone()
    if row is None:
        raise TraceNotFoundError(trace_id)
    return row


def list_traces_from_db(db_path: Path, count=500, reverse=False):
    list_sql = """
    SELECT id, created_at, COALESCE(LENGTH(msgpack), LENGTH(CAST(data AS BLOB)))
    FROM traces ORDER BY id DESC LIMIT ?
    """

    with db_connection(db_path) as connection:
        cursor = connection.execute(list_sql, [count])
        rows = cursor.fetchall()
    if reverse:
        return reversed(rows)
    return rows


def delete_traces_by_id(db_path: Path, trace_ids: Tuple[str, ...]):
    params = ", ".join("?" * len(trace_ids))
    delete_sql = f"DELETE FROM traces WHERE id in ({params})"

    with db_connection(db_path) as connection:
        connection.execute(delete_sql, trace_ids)


def delete_traces_before(db_path: Path, before: datetime):
    delete_sql = "DELETE FROM traces WHERE (created_at < ?)"

    with db_connection(db_path) as connection:
        connection.execute(delete_sql, (before,))
        cursor = connection.execute("SELECT changes()")
        deleted_count = cursor.fetchone()[0]
    return deleted_count


def vacuum_db(db_path):
    with db_connection(db_path) as connection:
        connection.execute("VACUUM")


def create_schema_table(connection) -> None:
    create_table_query = """
    CREATE TABLE IF NOT EXISTS schemas (
        id integer PRIMARY KEY,
        created_at TEXT DEFAULT (STRFTIME('%Y-%m-%d %H:%M:%f', 'NOW')) NOT NULL,
        git_commit TEXT NULL,
        data text NOT NULL
    );
    """

    connection.execute(create_table_query)


def save_schema(connection, schema, commit_sha) -> None:
    insert_sql = 'INSERT INTO schemas("data", "git_commit") VALUES(?, ?)'
    connection.execute(insert_sql, (json.dumps(schema), commit_sha))


def load_schema(db_path: Path) -> Tuple[Dict[str, Any], str]:
    load_sql = """SELECT data, git_commit FROM schemas
        ORDER BY created_at DESC LIMIT 1"""

    with db_connection(db_path) as connection:
        cursor = connection.execute(load_sql)
        data, commit = cursor.fetchone()

    data = json.loads(data)
    return data, commit


def load_schema_for_commit_sha(
    db_path: Path, commit_sha: str
) -> Tuple[Dict[str, Any], str]:
    load_sql = """SELECT data FROM schemas WHERE git_commit = ?
        ORDER BY created_at DESC LIMIT 1"""

    with db_connection(db_path) as connection:
        try:
            cursor = connection.execute(load_sql, [commit_sha])
        except sqlite3.OperationalError:
            row = None
        else:
            row = cursor.fetchone()

    if not row:
        raise SchemaNotFoundError(commit_sha)

    return json.loads(row[0])



def pretty_byte_size(size_bytes):
    if size_bytes < 1024:
        return f"{size_bytes:5d} B"

    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    bit_length = size_bytes.bit_length() - 1
    index = bit_length // 10
    denominator = 1 << (10 * index)
    ratio = size_bytes / denominator
    return f"{ratio:5.1f} {size_name[index]}"



from __future__ import annotations

import os
import threading
import time
import types
from typing import TYPE_CHECKING, Dict, List, Literal, Tuple

import ulid

from ..serialize import decode_body
from .types import BaseOutboundRequest, BaseOutboundResponse


class OutboundRequest(BaseOutboundRequest):
    subtype: Literal["requests"]


class OutboundResponse(BaseOutboundResponse):
    subtype: Literal["requests"]


class ApiRequestFilter:
    co_names: Tuple[str, ...] = ("send",)
    requests_filename = os.path.normpath("requests/sessions")

    def __init__(self, config) -> None:
        self.config = config
        self._frame_ids: Dict[int, str] = {}

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        filepath = frame.f_code.co_filename
        callable_name = frame.f_code.co_name
        return callable_name == "send" and self.requests_filename in filepath

    def process(
        self,
        frame: types.FrameType,
        event: str,
        arg: object,
        call_frames: List[Tuple[types.FrameType, str]],
    ):
        timestamp = time.time()
        thread = threading.current_thread()
        frame_locals = frame.f_locals
        request = frame_locals["request"]
        method_and_url = f"{request.method} {request.url}"

        if event == "call":
            frame_id = f"frm_{ulid.new()}"
            self._frame_ids[id(frame)] = frame_id

            api_request: OutboundRequest = {
                "body": decode_body(request.body, request.headers),
                "frame_id": frame_id,
                "headers": dict(request.headers),
                "method": request.method,
                "method_and_full_url": method_and_url,
                "subtype": "requests",
                "thread": thread.name,
                "thread_native_id": thread.native_id,
                "timestamp": timestamp,
                "type": "outbound_http_request",
                "url": request.url,
            }
            return api_request

        assert event == "return"

        response = arg
        if TYPE_CHECKING:
            from requests.models import Response

            assert isinstance(response, Response)

        if response is None:
            body = None
            headers = None
            status_code = None
        else:
            body = response.text
            headers = dict(response.headers)
            status_code = response.status_code

        api_response: OutboundResponse = {
            "body": body,
            "frame_id": self._frame_ids[id(frame)],
            "headers": headers,
            "method": request.method,
            "method_and_full_url": method_and_url,
            "status_code": status_code,
            "subtype": "requests",
            "thread": thread.name,
            "thread_native_id": thread.native_id,
            "timestamp": timestamp,
            "type": "outbound_http_response",
            "url": request.url,
        }
        return api_response



from __future__ import annotations

import os
import threading
import time
from typing import TYPE_CHECKING, Any, Dict, Literal, Optional, Tuple, Type, TypedDict

import ulid

if TYPE_CHECKING:
    from huey.api import Task


class HueyJob(TypedDict, total=False):
    frame_id: str
    name: str
    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]
    thread: str
    thread_native_id: Optional[int]
    timestamp: float
    type: Literal["background_job", "background_job_end"]
    subtype: Literal["huey"]


class HueyFilter:
    co_names: Tuple[str, ...] = ("execute",)
    huey_filename = os.path.normpath("/huey/api.py")
    klass: Type[Task] | None = None

    def __init__(self, config) -> None:
        self.config = config
        self._frame_ids: Dict[int, str] = {}

    def __call__(self, frame, event, arg):
        if (
            self.huey_filename in frame.f_code.co_filename
            and frame.f_code.co_name in self.co_names
        ):
            # Doing the import once and binding is substantially faster than going
            # through the import system every time.
            # See:
            # https://github.com/django/asgiref/issues/269
            # https://github.com/django/asgiref/pull/288
            # A more battle-tested (slightly slower/more complex) alternative would be:
            # https://github.com/django/django/pull/14850
            # https://github.com/django/django/pull/14858
            # https://github.com/django/django/pull/14931
            if HueyFilter.klass is None:
                from huey.api import Task

                HueyFilter.klass = Task
            return isinstance(frame.f_locals["self"], HueyFilter.klass)
        return False

    def process(self, frame, event, arg, call_frames):
        huey_job: HueyJob
        timestamp = time.time()
        thread = threading.current_thread()

        frame_locals = frame.f_locals
        task_object = frame_locals["self"]
        task_args, task_kwargs = task_object.data

        if event == "call":
            frame_id = f"frm_{ulid.new()}"
            self._frame_ids[id(frame)] = frame_id
            huey_job = {
                "frame_id": frame_id,
                "name": f"{task_object.__module__}.{task_object.name}",
                "args": task_args,
                "kwargs": task_kwargs,
                "thread": thread.name,
                "thread_native_id": thread.native_id,
                "timestamp": timestamp,
                "type": "background_job",
                "subtype": "huey",
            }
            return huey_job

        assert event == "return"

        huey_job = {
            "frame_id": self._frame_ids[id(frame)],
            "name": f"{task_object.__module__}.{task_object.name}",
            "args": task_args,
            "kwargs": task_kwargs,
            "thread": thread.name,
            "thread_native_id": thread.native_id,
            "timestamp": timestamp,
            "type": "background_job_end",
            "subtype": "huey",
        }
        return huey_job



from __future__ import annotations

import logging
import mimetypes
import os
import sqlite3
import sys
import threading
from itertools import chain
from typing import Awaitable, Callable

from asgiref.sync import iscoroutinefunction
from django.conf import settings
from django.http import (
    FileResponse,
    HttpRequest,
    HttpResponse,
    HttpResponseNotFound,
    JsonResponse,
)

from .config import load_config
from .db import list_traces_from_db, setup_db
from .profiler import KoloProfiler
from .serialize import monkeypatch_queryset_repr
from .web.home import kolo_web_home

logger = logging.getLogger("kolo")

DjangoView = Callable[[HttpRequest], HttpResponse]
DjangoAsyncView = Callable[[HttpRequest], Awaitable[HttpResponse]]


class KoloMiddleware:
    sync_capable = True
    async_capable = True

    def __init__(self, get_response: DjangoView | DjangoAsyncView) -> None:
        self._is_coroutine = iscoroutinefunction(get_response)
        self._get_response = get_response
        self.config = load_config()
        self.enabled = self.should_enable()
        if self.enabled:
            self.db_path = setup_db()

            # TODO: Put the full URL here not just the /_kolo/ path
            if self.config.get("web_experience", False):
                print("\nView recent requests at /_kolo/")

    def __call__(self, request: HttpRequest) -> HttpResponse:
        if self._is_coroutine:
            # TODO: Handle _kolo/ for asgi???
            get_response = self.aget_response
        else:
            if request.path.startswith("/_kolo/"):
                return kolo_web_router(request)

            get_response = self.get_response  # type: ignore

        # WARNING: Because Django's runserver uses threading, we need
        # to be careful about thread safety here.
        if not self.enabled or self.check_for_third_party_profiler():
            return get_response(request)  # type: ignore

        filter_config = self.config.get("filters", {})
        ignore_request_paths = filter_config.get("ignore_request_paths", [])
        for path in ignore_request_paths:
            if path in request.path:
                return get_response(request)  # type: ignore

        # Don't store the KoloProfiler on self to avoid threadsafety
        # bugs. If a different thread gets this instance of KoloProfiler
        # at the wrong time, we lose the original profiler's trace.
        profiler = KoloProfiler(
            self.db_path, config=self.config, source="kolo.middleware.KoloMiddleware"
        )

        monkeypatch_queryset_repr()
        if self._is_coroutine:
            return self.aprofile_response(request, profiler)
        else:
            return self.profile_response(request, profiler)

    def profile_response(self, request, profiler):
        with profiler:
            response = self.get_response(request)
        self.save_trace(profiler)
        return response

    async def aprofile_response(self, request, profiler):
        with profiler:
            response = await self.aget_response(request)
        self.save_trace(profiler)
        return response

    async def aget_response(self, request: HttpRequest) -> HttpResponse:
        response = await self._get_response(request)  # type: ignore
        return response

    def get_response(self, request: HttpRequest) -> HttpResponse:
        response = self._get_response(request)
        return response  # type: ignore

    @staticmethod
    def save_trace(profiler):
        name = "kolo-save_request_in_db"
        threading.Thread(target=profiler.save_request_in_db, name=name).start()

    def check_for_third_party_profiler(self) -> bool:
        profiler = sys.getprofile()
        if profiler:
            logger.warning("Profiler %s is active, disabling KoloMiddleware", profiler)
            return True
        if self.config.get("threading", False):
            try:
                profiler = threading.getprofile()  # type: ignore[attr-defined]
            except AttributeError:
                profiler = threading._profile_hook
            if profiler:  # pragma: no branch
                logger.warning(
                    "Profiler %s is active, disabling KoloMiddleware", profiler
                )
                return True
        return False

    def should_enable(self) -> bool:
        if settings.DEBUG is False:
            logger.debug("DEBUG mode is off, disabling KoloMiddleware")
            return False

        if os.environ.get("KOLO_DISABLE", "false").lower() not in ["false", "0"]:
            logger.debug("KOLO_DISABLE is set, disabling KoloMiddleware")
            return False

        if self.check_for_third_party_profiler():
            return False

        return True


def kolo_web_router(request: HttpRequest) -> HttpResponse:
    path = request.path
    if path == "/_kolo/":
        return kolo_web_home(request)
    elif path.startswith("/_kolo/static/"):
        static_dir = os.path.join(os.path.dirname(__file__), "web", "static")
        file_path = os.path.join(static_dir, request.path[len("/_kolo/static/") :])

        if os.path.exists(file_path):
            mime_type, encoding = mimetypes.guess_type(file_path)
            return FileResponse(  # type: ignore # seems iffy and difficult to fix with little benefit
                open(file_path, "rb"),
                content_type=mime_type or "application/octet-stream",
            )
        else:
            return HttpResponseNotFound("File not found")
    elif path.startswith("/_kolo/api/generate-test/"):
        return kolo_web_api_generate_test(request)
    elif path.startswith("/_kolo/api/traces/"):
        return kolo_web_api_get_trace(request)
    elif path.startswith("/_kolo/api/latest-traces/"):
        return kolo_web_api_latest_traces(request)
    else:
        return HttpResponseNotFound("Kolo Web: Not Found")


def kolo_web_api_generate_test(request: HttpRequest) -> HttpResponse:
    trace_id = request.path.replace("/_kolo/api/generate-test/", "").replace("/", "")

    from .generate_tests import generate_from_trace_id

    test_code = generate_from_trace_id(trace_id, "MyTestCase", "test_my_view")

    return JsonResponse({"test_code": test_code})


def kolo_web_api_get_trace(request: HttpRequest) -> HttpResponse:
    trace_id = request.path.replace("/_kolo/api/traces/", "").replace("/", "")

    from .db import load_trace_from_db

    db_path = setup_db()

    msgpack_data, json_data = load_trace_from_db(db_path, trace_id)
    if json_data:
        return HttpResponse(json_data, content_type="application/json")
    else:
        return HttpResponse(msgpack_data, content_type="application/msgpack")


def kolo_web_api_latest_traces(request: HttpRequest) -> HttpResponse:
    from .db import db_connection

    db_path = setup_db()

    with db_connection(db_path) as connection:
        cursor = connection.execute("SELECT id FROM traces ORDER BY id desc LIMIT 10")

        rows = cursor.fetchall()

        traces = list(chain.from_iterable(rows))

    return JsonResponse({"traces": traces})



from __future__ import annotations

import platform
from datetime import datetime, timezone

from jinja2 import ChoiceLoader, Environment, FileSystemLoader, PackageLoader

from ..config import load_config
from ..db import load_trace, load_trace_from_db, setup_db
from ..version import __version__
from .processors import load_processors, run_processors


class KoloPackageLoader(PackageLoader):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        # Work around UNC path mishandling:
        # https://github.com/pallets/jinja/issues/1675
        if platform.system() == "Windows":
            unc_prefix = "\\\\?\\"
            if self._template_root.startswith(unc_prefix):  # pragma: no cover
                self._template_root = self._template_root[len(unc_prefix) :]


env = Environment(
    loader=ChoiceLoader(
        (
            FileSystemLoader(""),
            KoloPackageLoader("kolo"),
        )
    )
)


def maybe_black(rendered):
    try:
        from black import format_file_contents
        from black.mode import Mode
        from black.parsing import InvalidInput
        from black.report import NothingChanged
    except ImportError:  # pragma: no cover
        return rendered

    try:
        return format_file_contents(
            rendered, fast=True, mode=Mode(magic_trailing_comma=False)
        )
    except (InvalidInput, NothingChanged):  # pragma: no cover
        return rendered


def generate_from_trace_id(
    trace_id: str,
    test_class: str,
    test_name: str,
    template_name: str = "",
    config=None,
    include_generation_timestamp=True,
) -> str:
    if config is None:
        config = load_config()
    db_path = setup_db()
    msgpack_data, json_data = load_trace_from_db(db_path, trace_id)
    trace = load_trace(msgpack_data, json_data)

    processors = load_processors(config)

    context = {
        "_config": config,
        "_db_path": db_path,
        "_frames": trace["frames_of_interest"],
        "_trace": trace,
        "base_test_case": "TestCase",
        "kolo_version": __version__,
        "now": datetime.now(timezone.utc) if include_generation_timestamp else None,
        "test_class": test_class,
        "test_name": test_name,
        "trace_id": trace_id,
    }
    run_processors(processors, context)

    if not template_name:
        template_name = "django_request_test.py.j2"
    template = env.get_template(template_name)
    rendered = template.render(**context)
    return maybe_black(rendered)



from __future__ import annotations

import logging
import os
import sys
import threading
from typing import Awaitable, Callable

from asgiref.sync import iscoroutinefunction
from django.conf import settings
from django.http import HttpRequest, HttpResponse, HttpResponseNotFound, JsonResponse

from ..config import load_config
from ..db import list_traces_from_db, setup_db

logger = logging.getLogger("kolo")

DjangoView = Callable[[HttpRequest], HttpResponse]
DjangoAsyncView = Callable[[HttpRequest], Awaitable[HttpResponse]]


def kolo_web_home(request: HttpRequest) -> HttpResponse:
    cssBundleUri = "static/svelte-build/main.css"
    jsBundleUri = "static/svelte-build/main.js"
    html = f"""
  <head>
      <meta charset="UTF-8">
      <meta name="viewport" content="width=device-width, initial-scale=1.0">
      <title>Kolo</title>

      <link rel="icon" href="favicon.ico">

      <!-- For light mode -->
      <link rel="icon" sizes="16x16" href="static/favicons/favicon-dark-16x16.png" media="(prefers-color-scheme: light)">
      <link rel="icon" sizes="32x32" href="static/favicons/favicon-dark-32x32.png" media="(prefers-color-scheme: light)">

      <!-- For dark mode -->
      <link rel="icon" sizes="16x16" href="static/favicons/favicon-light-16x16.png" media="(prefers-color-scheme: dark)">
      <link rel="icon" sizes="32x32" href="static/favicons/favicon-light-32x32.png" media="(prefers-color-scheme: dark)">

      <link rel="stylesheet" href="static/main.css" />
      <link rel="stylesheet" href="{cssBundleUri}" />
  </head>
  <body>
  <div style="display: flex">
    <div id="app" style="width: 50%;"></div>
    <div id="viz" style="width: 50%; height: 100vh; border-left: 2px solid var(--vscode-textSeparator-foreground);"></div>
  </div>
"""

    html += f'\n\n<script type="text/javascript" src="static/main.js"></script>'
    html += f'\n\n<script type="text/javascript" src="{jsBundleUri}"></script>'

    html += "</body>"

    return HttpResponse(html)



from __future__ import annotations

import os
import threading
import time
import types
from typing import Dict, List, Literal, Tuple

import ulid

from ..serialize import decode_body, decode_header_value
from .types import BaseOutboundRequest, BaseOutboundResponse


class OutboundRequest(BaseOutboundRequest):
    subtype: Literal["urllib3"]


class OutboundResponse(BaseOutboundResponse):
    subtype: Literal["urllib3"]


def get_full_url(frame_locals) -> str:
    scheme = frame_locals["self"].scheme
    host = frame_locals["self"].host
    port = frame_locals["self"].port
    url = frame_locals["url"]
    return f"{scheme}://{host}:{port}{url}"


class Urllib3Filter:
    co_names: Tuple[str, ...] = ("urlopen",)
    urllib3_filename = os.path.normpath("urllib3/connectionpool")

    def __init__(self, config) -> None:
        self.config = config
        self._frame_ids: Dict[int, str] = {}

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        filepath = frame.f_code.co_filename
        callable_name = frame.f_code.co_name
        return callable_name == "urlopen" and self.urllib3_filename in filepath

    def process(
        self,
        frame: types.FrameType,
        event: str,
        arg: object,
        call_frames: List[Tuple[types.FrameType, str]],
    ):
        frame_locals = frame.f_locals
        thread = threading.current_thread()

        if event == "call":
            request_headers = {
                key: decode_header_value(value)
                for key, value in frame_locals["headers"].items()
            }

            full_url = get_full_url(frame_locals)
            method = frame_locals["method"].upper()
            method_and_full_url = f"{method} {full_url}"

            frame_id = f"frm_{ulid.new()}"
            self._frame_ids[id(frame)] = frame_id
            api_request: OutboundRequest = {
                "body": decode_body(frame_locals["body"], request_headers),
                "frame_id": frame_id,
                "headers": request_headers,
                "method": method,
                "method_and_full_url": method_and_full_url,
                "subtype": "urllib3",
                "thread": thread.name,
                "thread_native_id": thread.native_id,
                "timestamp": time.time(),
                "type": "outbound_http_request",
                "url": full_url,
            }
            return api_request

        assert event == "return"

        full_url = get_full_url(frame_locals)
        method = frame_locals["method"].upper()
        method_and_full_url = f"{method} {full_url}"

        try:
            response = frame_locals["response"]
        except KeyError:
            headers = None
            status = None
        else:
            headers = dict(response.headers)
            status = response.status

        api_response: OutboundResponse = {
            "body": None,
            "frame_id": self._frame_ids[id(frame)],
            "headers": headers,
            "method": method,
            "method_and_full_url": method_and_full_url,
            "status_code": status,
            "subtype": "urllib3",
            "thread": thread.name,
            "thread_native_id": thread.native_id,
            "timestamp": time.time(),
            "type": "outbound_http_response",
            "url": full_url,
        }
        return api_response



import json
import re
from ast import literal_eval
from collections import Counter
from datetime import datetime, timedelta, timezone
from decimal import Decimal
from functools import total_ordering
from textwrap import dedent
from typing import Dict, List, Set, Union

import sqlglot
from more_itertools import unique_everseen

from kolo.django_schema import verbose_name

from .sort import TopologicalSorter

DATABASES = {
    "postgresql": "postgres",
}


UUID_REPR_REGEX = re.compile(
    r"UUID\(\'([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})\'\)"
)
DECIMAL_REPR_REGEX = re.compile(r"Decimal\('(\d+.?\d+?)'\)")
TIMEDELTA_REPR_REGEX = re.compile(
    r"datetime.timedelta\((?:days=(\d+),? ?)?(?:seconds=(\d+),? ?)?(?:microseconds=(\d+))?\)"
)
TIMEDELTA_STR_REGEX = re.compile(r"(\d+) days (\d+)\.(\d+) seconds")
TIMEDELTA_MICROSECONDS_REGEX = re.compile(r"\d+")


def is_auto_now(field):
    django_field = field.schema["django_field"]
    if django_field not in (
        "django.db.models.fields.DateTimeField",
        "django.db.models.fields.DateField",
    ):
        return False
    return field.schema["auto_now"] or field.schema["auto_now_add"]


def parse_value(value, column_schema):
    if value is None:
        return None
    django_field = column_schema["django_field"]
    if django_field == "django.db.models.fields.UUIDField":
        uuid_match = UUID_REPR_REGEX.match(str(value))
        if uuid_match:
            value = uuid_match[1]
    elif django_field == "django.db.models.fields.json.JSONField":
        try:
            value = json.loads(value)
        except json.decoder.JSONDecodeError:
            pass
    elif django_field == "django.db.models.fields.DecimalField":
        decimal_match = DECIMAL_REPR_REGEX.match(str(value))
        if decimal_match:
            value = Decimal(decimal_match[1])
    elif django_field == "django.db.models.fields.DurationField":
        if isinstance(value, str):
            timedelta_match = TIMEDELTA_REPR_REGEX.match(value)
            duration_match = TIMEDELTA_STR_REGEX.match(value)
            microseconds_match = TIMEDELTA_MICROSECONDS_REGEX.match(value)
            if timedelta_match:
                groups = [int(g) if g else 0 for g in timedelta_match.groups()]
                value = timedelta(
                    days=groups[0], seconds=groups[1], microseconds=groups[2]
                )
            elif duration_match:
                groups = duration_match.groups()  # type: ignore
                value = timedelta(
                    days=int(groups[0]),
                    seconds=int(groups[1]),
                    microseconds=int(groups[2]),
                )
            elif microseconds_match:  # pragma: no branch
                value = timedelta(microseconds=int(microseconds_match.group()))
        else:
            value = timedelta(microseconds=value)
    elif django_field == "django.db.models.fields.DateTimeField":
        if isinstance(value, str):
            value = datetime.fromisoformat(value)
        value = datetime.combine(
            value.date(), value.time(), value.tzinfo or timezone.utc
        )
    return value


class TooComplicatedError(Exception):
    pass


class AssertInsert:
    def __init__(self, variable_name, model, fields, import_path=None, query=None):
        self.variable_name = variable_name
        self.model = model
        self.fields = fields
        self.query = query
        self.import_path = import_path

    @classmethod
    def from_raw(cls, fields, query, table_schema, names, primary_key):
        module = table_schema["model_module"]
        model = table_schema["model_name"]
        verbose_name = table_schema["verbose_name"]
        variable_name = names.next_variable_name(
            verbose_name, primary_key.name, primary_key.value_repr
        )
        return cls(
            variable_name, f"{model}", fields, f"from {module} import {model}", query
        )

    def update_fields(self, names, schema):
        fields = []
        for field in self.fields:
            if field.schema["primary_key"]:
                continue  # pragma: no cover
            if field.schema["django_field"] == "django.db.models.fields.DateTimeField":
                continue  # pragma: no cover
            if field.schema["django_field"] == "django.db.models.fields.json.JSONField":
                continue  # pragma: no cover
            if field.skip_default():
                continue

            if field.schema["is_relation"]:
                related_model = field.schema["related_model"]
                if related_model is None:
                    fields.append(field)
                    continue
                verbose_name = schema[related_model]["verbose_name"]
                related_pk = field.schema["related_pk"]
                try:
                    field.value_repr = field.value = names.names[
                        (verbose_name, related_pk, field.value_repr)
                    ]
                except KeyError:  # pragma: no cover
                    pass

            fields.append(field)
        self.fields = fields

    def __eq__(self, other):
        if not isinstance(other, AssertInsert):
            return NotImplemented  # pragma: no cover
        return (self.variable_name, self.model, self.fields) == (
            other.variable_name,
            other.model,
            other.fields,
        )

    def __repr__(self):
        return dedent(  # pragma: no cover
            f"""\
            AssertInsert(
                "{self.variable_name}",
                "{self.model}",
                {self.fields},
            )"""
        )

    def unique_fields(self):
        unique = []
        rest = []
        for field in self.fields:
            if field.schema.get("has_unique", False):
                unique.append(field)
            else:
                rest.append(field)
        return unique, rest

    def foreign_keys(self):
        foreign = []
        rest = []
        for field in self.fields:
            if field.schema["is_relation"]:
                foreign.append(field)
            else:
                rest.append(field)
        return foreign, rest

    def get_template_parts(self, fields):
        yield f"{self.variable_name} = {self.model}.objects.get("
        for field in fields:
            yield f"{field.name}={field.value_repr},"
        yield ")"

    def get_fields(self):
        fields, rest = self.unique_fields()
        if fields:
            return fields, rest

        fields, rest = self.foreign_keys()
        if fields:
            return fields, rest

        return self.fields, []

    @property
    def template_parts(self):
        fields, rest = self.get_fields()

        yield from self.get_template_parts(fields)

        for field in rest:
            yield field.assert_equal_template(self.variable_name)

    @property
    def pytest_template_parts(self):
        fields, rest = self.get_fields()

        yield from self.get_template_parts(fields)

        for field in rest:
            yield field.assert_equal_template_pytest(self.variable_name)


class AssertUpdate:
    def __init__(self, variable_name, model, table, fields, query=None):
        self.variable_name = variable_name
        self.model = model
        self.table = table
        self.fields = fields
        self.query = query

    def __repr__(self):
        return dedent(
            f"""\
            AssertUpdate(
                "{self.variable_name}",
                "{self.model}",
                "{self.table}",
                {self.fields},
            )"""
        )

    def __eq__(self, other):
        if not isinstance(other, AssertUpdate):
            return NotImplemented  # pragma: no cover
        return (self.variable_name, self.fields) == (other.variable_name, other.fields)

    def update_fields(self, names, schema):
        fields = []
        for field in self.fields:
            if field.schema["primary_key"]:
                continue  # pragma: no cover
            if field.schema["django_field"] == "django.db.models.fields.DateTimeField":
                continue  # pragma: no cover
            if field.schema["django_field"] == "django.db.models.fields.json.JSONField":
                continue  # pragma: no cover

            if field.schema["is_relation"]:
                related_model = field.schema["related_model"]
                verbose_name = schema[related_model]["verbose_name"]
                related_pk = field.schema["related_pk"]
                try:
                    field.value_repr = field.value = names.names[
                        (verbose_name, related_pk, field.value_repr)
                    ]
                except KeyError:  # pragma: no cover
                    pass
                else:
                    field.name = field.schema["field_name"]

            fields.append(field)

        for name, schema in schema[self.table]["fields"].items():
            if schema["gfk_object_id_field"]:
                fields = collect_gfk(fields, name, schema, names)
        self.fields = fields

    @property
    def template_parts(self):
        if not self.fields:
            return
        yield f"{self.variable_name}.refresh_from_db()"
        for field in self.fields:
            yield field.assert_equal_template(self.variable_name)

    @property
    def pytest_template_parts(self):
        if not self.fields:
            return
        yield f"{self.variable_name}.refresh_from_db()"
        for field in self.fields:
            yield field.assert_equal_template_pytest(self.variable_name)


class AssertDelete:
    def __init__(self, model, fields, query=None):
        self.model = model
        self.fields = fields
        self.query = query

    def __repr__(self):
        return dedent(
            f"""\
            AssertDelete(
                "{self.model}",
                {self.fields},
            )"""
        )

    def __eq__(self, other):
        if not isinstance(other, AssertDelete):
            return NotImplemented  # pragma: no cover
        return (self.model, self.fields) == (other.model, other.fields)

    def update_fields(self, names, schema):
        pass

    @property
    def template_parts(self):
        yield f"self.assertFalse({self.model}.objects.filter("
        for name, fields in self.fields.items():
            if len(fields) == 1:
                yield f"{name}={fields[0].value_repr},"
            else:
                yield f"{name}__in=("
                for field in fields:
                    yield f"{field.value_repr},"
                yield "),"
        yield ").exists())"

    @property
    def pytest_template_parts(self):
        yield f"assert not {self.model}.objects.filter("
        for name, fields in self.fields.items():
            if len(fields) == 1:
                yield f"{name}={fields[0].value_repr},"
            else:
                yield f"{name}__in=("
                for field in fields:
                    yield f"{field.value_repr},"
                yield "),"
        yield ").exists()"


class DjangoCreate:
    def __init__(
        self,
        table,
        model,
        primary_key,
        values,
        variable_name,
        import_path=None,
        query=None,
    ):
        self.table = table
        self.model = model
        self.primary_key = primary_key
        self.values = values
        self.import_path = import_path
        self.query = query
        self.variable_name = variable_name

    @classmethod
    def from_raw(cls, table, values, query, schema_data, names):
        module = schema_data[table]["model_module"]
        model = schema_data[table]["model_name"]
        values = tuple(values)
        primary_key = find_pk(values)
        verbose_name = schema_data[table]["verbose_name"]
        variable_name = names.next_variable_name(
            verbose_name, primary_key.name, primary_key.value_repr
        )
        return cls(
            table,
            f"{model}",
            primary_key,
            values,
            variable_name,
            f"from {module} import {model}",
            query,
        )

    def update_fields(self, names, schema):
        fields = self.values
        for name, field_schema in schema[self.table]["fields"].items():
            if field_schema["gfk_object_id_field"]:
                fields = collect_gfk(fields, name, field_schema, names)

        updated_fields = []
        for field in fields:
            if field.schema["primary_key"]:
                continue
            if field.schema["null"] and field.value is None:
                continue
            if is_auto_now(field):
                continue
            if field.skip_default():
                continue
            if field.schema["is_relation"]:
                related_model = field.schema["related_model"]
                if related_model is None:
                    updated_fields.append(field)
                    continue  # Generic Foreign Key
                verbose_name = schema[related_model]["verbose_name"]
                related_pk = field.schema["related_pk"]
                try:
                    field.value_repr = field.value = names.names[
                        (verbose_name, related_pk, field.value_repr)
                    ]
                except KeyError:
                    field.name = field.schema["field_attname"]
            updated_fields.append(field)

        self.values = tuple(updated_fields)

    def __eq__(self, other):
        if not isinstance(other, DjangoCreate):
            return NotImplemented
        return (self.table, self.model, self.values, self.variable_name) == (
            other.table,
            other.model,
            other.values,
            other.variable_name,
        )

    def __repr__(self):
        return dedent(
            f"""\
            DjangoCreate(
                "{self.table}",
                "{self.model}",
                {self.primary_key},
                {self.values},
                "{self.variable_name}",
            )"""
        )

    @property
    def template_parts(self):
        yield f"{self.variable_name}, _created = {self.model}.objects.get_or_create("
        for field in self.values:
            yield f"{field.name}={field.value_repr},"
        defaults = {self.primary_key.name: self.primary_key.raw_value}
        yield f"defaults={defaults},"
        yield ")"


@total_ordering
class DjangoField:
    def __init__(self, name, column_name, value, schema=None):
        self.name = name
        self.column_name = column_name
        self.value_repr = repr(value)
        self.raw_value = value
        self.value = repr(value) if isinstance(value, str) else value
        self.schema = schema

    @classmethod
    def from_column(cls, column, value, table_schema):
        field_schema = table_schema["fields"][column.name]
        name = field_schema["field_name"]
        value = parse_value(value, field_schema)
        return cls(name, column.name, value, field_schema)

    @classmethod
    def from_identifier(cls, identifier, value, table_schema):
        field_schema = table_schema["fields"][identifier.name]
        if isinstance(value, sqlglot.expressions.Null):
            value = None
        elif isinstance(value, sqlglot.expressions.Boolean):
            value = value.this
        elif isinstance(value, sqlglot.expressions.Cast):
            value = value.name
        elif isinstance(value, sqlglot.expressions.Anonymous):
            value = str(value)
        elif value.is_string:
            value = value.name
        else:
            value = literal_eval(value.name)
        value = parse_value(value, field_schema)
        return cls(field_schema["field_name"], identifier.name, value, field_schema)

    @classmethod
    def from_eq(cls, eq, schema_fields):
        name = eq.left.name
        right = eq.right
        if isinstance(right, sqlglot.expressions.Literal):
            value = right.name
        elif isinstance(right, sqlglot.expressions.Boolean):
            value = right.this
        elif isinstance(right, sqlglot.expressions.Cast) and right.name:
            value = right.name
            # TODO: Do something reasonable with `right.to`
        elif isinstance(right, sqlglot.expressions.Paren):
            raise TooComplicatedError()
        else:
            raise TooComplicatedError()  # pragma: no cover
        field_schema = schema_fields[name]
        value = parse_value(value, field_schema)
        return cls(field_schema["field_attname"], name, value, field_schema)

    def __eq__(self, other):
        return (self.name, self.column_name, self.value) == (
            other.name,
            other.column_name,
            other.value,
        )

    def __lt__(self, other) -> bool:
        if self.name == other.name:
            if self.value is None:
                return False
            if other.value is None:
                return True
            return repr(self.value) < repr(other.value)
        return self.name < other.name

    def __repr__(self):
        return f'DjangoField("{self.name}", "{self.column_name}", {self.value})'

    def assert_equal_template(self, variable_name):
        return f"self.assertEqual({variable_name}.{self.name}, {self.value_repr})"

    def assert_equal_template_pytest(self, variable_name):
        return f"assert {variable_name}.{self.name} == {self.value_repr}"

    def skip_default(self):
        if "default" not in self.schema:
            return False

        default_schema = self.schema["default"]
        if repr(default_schema["value"]) == self.value_repr:
            return True

        if "type" not in default_schema:
            return False

        field_type = default_schema["type"]
        value = default_schema["value"]
        if field_type == "Decimal" and self.value == Decimal(value):
            return True
        if field_type == "timedelta" and self.value == timedelta(  # pragma: no branch
            **value
        ):
            return True
        return False


class VariableNames:
    def __init__(self):
        self.counts: Counter = Counter()
        self.names = {}

    def next_variable_name(self, verbose_name, field_name, field_repr):
        full_key = verbose_name, field_name, field_repr
        try:
            return self.names[full_key]
        except KeyError:
            pass

        key = verbose_name, field_name
        self.counts[key] += 1
        suffix = self.counts[key]
        if suffix == 1:
            name = verbose_name
        else:
            name = f"{verbose_name}_{suffix}"
        self.names[full_key] = name
        return name


def build_insert_assert_fields(columns, table_schema, names):
    fields = []
    for identifier, value in columns:
        field = DjangoField.from_identifier(identifier, value, table_schema)
        django_field = table_schema["fields"][field.column_name]["django_field"]
        if django_field == "django.db.models.fields.DateTimeField":
            continue
        fields.append(field)

    for name, schema in table_schema["fields"].items():
        if schema["gfk_object_id_field"]:
            fields = collect_gfk(fields, name, schema, names)
    return fields


def collect_gfk(fields, name, schema, names):
    try:
        from django.contrib.contenttypes.models import ContentType
    except ImportError:  # pragma: no cover
        return fields

    content_type_field = None
    object_id_field = None

    for field in fields:
        if field.schema["field_name"] == schema["gfk_content_type_field"]:
            content_type_field = field
        elif field.schema["field_name"] == schema["gfk_object_id_field"]:
            object_id_field = field

    if content_type_field is None or object_id_field is None:
        return fields  # pragma: no cover

    try:
        content_type = ContentType.objects.get(pk=content_type_field.raw_value)
    except ContentType.DoesNotExist:  # pragma: no cover
        return fields

    model = content_type.model_class()
    value = names.next_variable_name(
        verbose_name(model),
        model._meta.pk.attname,  # type: ignore
        object_id_field.raw_value,
    )
    gfk_field = DjangoField(
        name=name,
        column_name=name,
        value=value,
        schema=schema,
    )
    gfk_field.value_repr = gfk_field.value = value
    collected = []
    for field in fields:
        if field == content_type_field or field == object_id_field:
            continue
        collected.append(field)
    collected.append(gfk_field)
    return collected


def build_update_assert_fields(columns, schema_fields, names):
    fields = []
    for eq in columns:
        if not isinstance(eq, sqlglot.expressions.EQ):
            continue
        try:
            field = DjangoField.from_eq(eq, schema_fields)
        except TooComplicatedError:
            continue
        django_field = schema_fields[field.column_name]["django_field"]
        if django_field == "django.db.models.fields.DateTimeField":
            continue
        if field.skip_default():
            continue
        fields.append(field)
    return fields


def build_delete_lookup_fields(where, table_schema):
    fields: Dict[str, List[DjangoField]] = {}
    if isinstance(where.this, sqlglot.expressions.In):  # pragma: no branch
        for literal in where.this.expressions:
            field = DjangoField.from_identifier(where.this.this, literal, table_schema)
            fields.setdefault(field.name, []).append(field)
    return fields


def find_pk(fields):
    for field in fields:
        if field.schema["primary_key"]:
            return field
    return None  # pragma: no cover


def sorted_inserts(deduplicated, schema_data):
    objects = {
        (obj.table, obj.primary_key.column_name, obj.primary_key.raw_value): obj
        for obj in deduplicated
    }
    graph = []
    for obj in objects.values():
        schema = schema_data[obj.table]["fields"]
        keys = (
            (
                schema[field.column_name]["related_model"],
                schema[field.column_name]["related_pk"],
                field.raw_value,
            )
            for field in obj.values
            if schema[field.column_name]["is_relation"]
        )
        graph.append((obj, [objects[key] for key in keys if key in objects]))

    sorter = TopologicalSorter(graph)
    return list(sorter.static_order())


def parse_columns(columns, row, schema_data):
    columns_by_table: Dict[str, List[DjangoField]] = {}
    for column, value in zip(columns, row):
        if not isinstance(column, sqlglot.exp.Column):
            continue

        table_schema = schema_data[column.table]
        field = DjangoField.from_column(column, value, table_schema)
        columns_by_table.setdefault(column.table, []).append(field)
    return columns_by_table


AnyAssert = Union[AssertDelete, AssertInsert, AssertUpdate]


def parse_sql_queries(sql_queries, schema_data):
    names = VariableNames()
    # each select needs to become an insert
    inserts = []
    imports = set()
    asserts: List[AnyAssert] = []
    seen_mutations: Dict[str, Set[str]] = {}
    for query in sql_queries:
        if query["query_data"] is None or query["query"] is None:
            continue

        database = query["database"]
        database = DATABASES.get(database, database)
        try:
            parsed_query = sqlglot.parse_one(query["query"], read=database)
        except sqlglot.errors.ParseError as e:  # pragma: no cover
            print("# Error parsing query:")
            print(f'# {query["query"]}')
            print(f"# {e.errors}")
            continue

        if isinstance(parsed_query, sqlglot.exp.Select):
            columns = [column for (_key, column) in parsed_query.iter_expressions()]
            for batch in query["query_data"]:
                # Skip queries with no columns, eg .count(), .exists()
                if not any(
                    isinstance(column, sqlglot.exp.Column) for column in columns
                ):
                    continue
                for row in batch:
                    columns_by_table = parse_columns(columns, row, schema_data)

                    for table, row in columns_by_table.items():
                        pk = find_pk(row)
                        if pk is None:
                            continue
                        elif pk.value_repr in seen_mutations.setdefault(table, set()):
                            continue
                        else:
                            create = DjangoCreate.from_raw(
                                table, row, query, schema_data, names
                            )
                            if create.primary_key.raw_value is None:
                                continue
                            inserts.append(create)
                            imports.add(create.import_path)
        else:
            _table = parsed_query.find(sqlglot.exp.Table)
            if _table is None:
                continue  # pragma: no cover
            table_name = _table.name
            table_schema = schema_data[table_name]
            columns = [column for (_key, column) in parsed_query.iter_expressions()]
            if isinstance(parsed_query, sqlglot.expressions.Update):
                where = parsed_query.find(sqlglot.expressions.Where)
                if where is None:
                    continue  # pragma: no cover

                fields = build_update_assert_fields(
                    columns, table_schema["fields"], names
                )

                if isinstance(where.this, sqlglot.expressions.In):
                    lookups = []
                    lookup_name = where.this.this.name
                    for literal in where.this.expressions:
                        lookups.append((lookup_name, literal.name))
                elif isinstance(where.this, sqlglot.expressions.Paren):
                    continue
                else:
                    lookup_name = where.this.left.name
                    lookup_value = where.this.right.name
                    lookups = [(lookup_name, lookup_value)]

                verbose_name = table_schema["verbose_name"]
                for lookup_name, lookup_value in lookups:
                    variable_name = names.next_variable_name(
                        verbose_name, lookup_name, lookup_value
                    )
                    assert_update = AssertUpdate(
                        variable_name,
                        table_schema["model_name"],
                        table_name,
                        fields,
                        query,
                    )
                    asserts.append(assert_update)
                    if table_schema["fields"][lookup_name][  # pragma: no branch
                        "primary_key"
                    ]:
                        seen_mutations.setdefault(table_name, set()).add(lookup_value)
            elif isinstance(parsed_query, sqlglot.expressions.Delete):
                where = parsed_query.find(sqlglot.expressions.Where)
                if where is None:
                    continue  # pragma: no cover

                model = table_schema["model_name"]
                module = table_schema["model_module"]
                fields = build_delete_lookup_fields(where, table_schema)
                assert_delete = AssertDelete(model, fields, query)
                asserts.append(assert_delete)
                import_path = f"from {module} import {model}"
                imports.add(import_path)
            elif isinstance(  # pragma: no branch
                parsed_query, sqlglot.expressions.Insert
            ):
                for assert_insert in build_assert_inserts(
                    parsed_query, query, table_schema, table_name, seen_mutations, names
                ):
                    asserts.append(assert_insert)
                    imports.add(assert_insert.import_path)

    deduplicated = unique_everseen(inserts)
    sql_fixtures = sorted_inserts(deduplicated, schema_data)
    for create in sql_fixtures:
        create.update_fields(names, schema_data)
    for assert_ in asserts:
        assert_.update_fields(names, schema_data)
    return sql_fixtures, imports, merge_asserts(asserts)


def build_assert_inserts(
    parsed_query, query, table_schema, table_name, seen_mutations, names
):
    sql_schema = parsed_query.find(sqlglot.expressions.Schema)
    if sql_schema is None:
        return  # pragma: no cover
    schema_columns = [
        column
        for (_key, column) in sql_schema.iter_expressions()
        if isinstance(column, sqlglot.expressions.Identifier)
    ]
    returning = parsed_query.find(sqlglot.expressions.Returning)
    values = parsed_query.find(sqlglot.expressions.Values)
    for sql_values_tuple in values.expressions:
        value_columns = [
            column for (_key, column) in sql_values_tuple.iter_expressions()
        ]
        zipped_columns = list(zip(schema_columns, value_columns))
        fields = build_insert_assert_fields(zipped_columns, table_schema, names)
        primary_key = None
        if returning:  # pragma: no branch
            for field, value in zip(returning.expressions, query["query_data"][0]):
                field_schema = table_schema["fields"][field.name]
                if field_schema["primary_key"]:  # pragma: no branch
                    primary_key = DjangoField(
                        field.this.name,
                        field.this.name,
                        value,
                        field_schema,
                    )
                    seen_mutations.setdefault(table_name, set()).add(repr(value))
        if primary_key is None:  # pragma: no cover
            continue
        yield AssertInsert.from_raw(fields, query, table_schema, names, primary_key)


def merge_asserts(asserts):
    lookup = {}
    merged: List[AnyAssert] = []
    for _assert in asserts:
        if isinstance(_assert, AssertInsert):
            lookup[_assert.variable_name] = _assert
            merged.append(_assert)
        elif isinstance(_assert, AssertUpdate):
            if _assert.variable_name not in lookup:
                merged.append(_assert)
                continue
            insert = lookup[_assert.variable_name]
            insert.fields = merge_fields(insert, _assert)
        else:
            merged.append(_assert)

    return tidy_deletes(merged)


def tidy_deletes(asserts):
    lookup: Dict[str, List[List[DjangoField]]] = {}
    tidied: List[AnyAssert] = []
    for _assert in reversed(asserts):
        if isinstance(_assert, AssertDelete):
            lookup.setdefault(_assert.model, []).append(_assert.fields)
            tidied.append(_assert)
        else:
            if _assert.model not in lookup:
                tidied.append(_assert)
                continue

            if any(
                would_be_deleted(_assert, delete_fields)
                for delete_fields in lookup[_assert.model]
            ):
                continue
            tidied.append(_assert)

    return list(reversed(tidied))


def would_be_deleted(_assert, delete_fields):
    if not any(field.name in delete_fields for field in _assert.fields):
        return False
    for field in _assert.fields:
        if field.name not in delete_fields:
            continue
        if field not in delete_fields[field.name]:
            return False
    return True


def merge_fields(insert, update):
    names = {f.name for f in update.fields}
    fields = []
    for field in insert.fields:
        if field.name in names:
            continue
        fields.append(field)
    fields.extend(update.fields)
    return fields



from urllib.parse import parse_qsl

from more_itertools import locate, peekable


def _format_header(header: str) -> str:
    header = header.upper().replace("-", "_")
    if header in ("CONTENT_LENGTH", "CONTENT_TYPE"):
        return header
    return f"HTTP_{header}"


def get_request_headers(request):
    request_headers = {
        _format_header(header): value for header, value in request["headers"].items()
    }
    if "HTTP_COOKIE" in request_headers and request_headers["HTTP_COOKIE"] == "":
        del request_headers["HTTP_COOKIE"]

    request_headers_to_delete = [
        "CONTENT_LENGTH",
        "HTTP_HOST",
        "HTTP_X_FORWARDED_FOR",
        "HTTP_X_FORWARDED_PROTO",
        "HTTP_CONNECTION",
        "HTTP_CACHE_CONTROL",
        "HTTP_DNT",
        "HTTP_SEC_CH_UA",
        "HTTP_USER_AGENT",
        "HTTP_ACCEPT",
        "HTTP_SEC_FETCH_DEST",
        "HTTP_SEC_CH_UA_MOBILE",
        "HTTP_REFERER",
        "HTTP_ACCEPT_ENCODING",
        "HTTP_ACCEPT_LANGUAGE",
        "HTTP_SEC_FETCH_SITE",
        "HTTP_SEC_FETCH_MODE",
        "HTTP_SEC_FETCH_USER",
        "HTTP_SEC_FETCH_DEST",
        "HTTP_SEC_CH_UA_PLATFORM",
        "HTTP_ORIGIN",
        "HTTP_UPGRADE_INSECURE_REQUESTS",
    ]

    for request_header in request_headers_to_delete:
        if request_header in request_headers:
            del request_headers[request_header]

    if "CONTENT_TYPE" in request_headers:
        # This is a special header, which ends up influencing
        # how the django test client formats different parts of the
        # request. It's provided to the test client as a lowercased
        # argument
        request_headers["content_type"] = request_headers["CONTENT_TYPE"]
        del request_headers["CONTENT_TYPE"]
    return request_headers


def get_request_body(request, request_headers):
    request_body = request["body"]
    content_type = request_headers.get("content_type", "")
    if content_type == "application/x-www-form-urlencoded":
        # So: How can we now format the body so that it becomes more readable
        # We can totally cater to just this urlencoded version first...

        # TODO: Eventually need to be able to support multivaluedicts / query params
        # with multiple values per key
        # I couldn't quite get that to work.
        # Custom request body formatting is clearly useful also here again..
        urldecoded_body = dict(parse_qsl(request_body))
        return f"urlencode({urldecoded_body})"
    elif "multipart/form-data" in content_type:
        request_headers.pop("content_type")
        if request["method"] == "POST":
            return str(request["post_data"])
        return ""
    else:
        return repr(request_body)


def get_query_params(request):
    query_params = request["query_params"]
    if query_params:
        return f"{query_params},"
    return ""


def group_frames(frames):
    request_indices = peekable(locate(frames, lambda f: f["type"] == "django_request"))
    response_indices = locate(frames, lambda f: f["type"] == "django_response")
    while True:
        try:
            request_index = next(request_indices)
            response_index = next(response_indices)
            while request_indices.peek(len(frames)) < response_index:
                request_index = next(request_indices)
            while response_index < request_index:
                response_index = next(response_indices)
        except StopIteration:
            break

        yield frames[request_index], frames[request_index + 1 : response_index], frames[
            response_index
        ]


def parse_request_frames(frames):
    served_request_frames = []
    for request, inner_frames, response in group_frames(frames):
        served_request_frames.append(
            {
                "request": request,
                "response": response,
                "templates": [
                    frame["template"]
                    for frame in inner_frames
                    if frame["type"] == "django_template_start"
                ],
                "frames": inner_frames,
            }
        )
    return served_request_frames


def build_test_client_call(
    request, query_params, prettified_request_body, request_headers, *, pytest
):
    client = "client" if pytest else "self.client"
    headers = "\n".join(
        f"{header}={repr(value)}," for header, value in request_headers.items()
    )
    method = request["method"].lower()
    path_info = request["path_info"]
    return f"""{client}.{method}(
            "{path_info}",
            {query_params}
            {prettified_request_body},
            {headers}
        )
    """



from datetime import datetime
from importlib import import_module
from itertools import chain

import click

from ..db import SchemaNotFoundError, load_schema_for_commit_sha
from ..django_schema import get_schema
from ..git import COMMIT_SHA
from .factories import build_factories, import_factories
from .outbound import parse_outbound_frames
from .queries import parse_sql_queries
from .request import (
    build_test_client_call,
    get_query_params,
    get_request_body,
    get_request_headers,
    parse_request_frames,
)

SCHEMA_NOT_FOUND_WARNING = """\
Warning: Could not find a Django schema for commit: {0}.

Falling back to the current commit's schema.

For more reliable results, run `kolo store-django-model-schema` from {0}
then retry this command.
"""


def process_django_schema(context):
    commit_sha = context["_trace"]["current_commit_sha"]
    db_path = context["_db_path"]
    if commit_sha == COMMIT_SHA:
        schema_data = get_schema()  # pragma: no cover
    else:
        try:
            schema_data = load_schema_for_commit_sha(db_path, commit_sha)
        except SchemaNotFoundError:
            click.echo(SCHEMA_NOT_FOUND_WARNING.format(commit_sha), err=True)
            schema_data = get_schema()
    return {"schema_data": schema_data}


def process_sql_queries(context):
    schema_data = context["schema_data"]

    all_imports = set()
    for section in context["sections"]:
        frames = section["frames"]
        sql_queries = [frame for frame in frames if frame["type"] == "end_sql_query"]
        sql_fixtures, imports, asserts = parse_sql_queries(sql_queries, schema_data)
        section["asserts"] = asserts
        section["sql_fixtures"] = sql_fixtures
        all_imports |= imports
    context["imports"] = sorted(all_imports)


def process_factories(context):
    imports = set(context.get("imports", []))
    for section in context["sections"]:
        sql_fixtures = section.get("sql_fixtures", [])
        factory_configs = import_factories(context["_config"])
        factories, factory_imports = build_factories(sql_fixtures, factory_configs)
        imports |= factory_imports
        section["sql_fixtures"] = factories
    context["imports"] = sorted(imports)


def process_django_request(context):
    frames = context["_frames"]
    served_request_frames = parse_request_frames(frames)

    sections = []
    if not served_request_frames:
        sections.append({"frames": frames})

    for served in served_request_frames:
        request = served["request"]
        response = served["response"]
        request_headers = get_request_headers(request)
        prettified_request_body = get_request_body(request, request_headers)
        query_params = get_query_params(request)
        request_timestamp = datetime.utcfromtimestamp(request["timestamp"]).isoformat(
            timespec="seconds"
        )
        template_names = served["templates"]

        test_client_call = build_test_client_call(
            request,
            query_params,
            prettified_request_body,
            request_headers,
            pytest=False,
        )
        test_client_call_pytest = build_test_client_call(
            request, query_params, prettified_request_body, request_headers, pytest=True
        )

        sections.append(
            {
                "request": request,
                "request_timestamp": request_timestamp,
                "response": response,
                "template_names": template_names,
                "test_client_call": test_client_call,
                "test_client_call_pytest": test_client_call_pytest,
                "frames": served["frames"],
            }
        )

    return {"sections": sections}


def process_outbound_requests(context):
    for section in context["sections"]:
        frames = section["frames"]
        outbound_request_frames = parse_outbound_frames(frames)
        section["outbound_request_frames"] = outbound_request_frames


def process_django_version(context):
    try:
        from django import __version__ as django_version
    except ImportError:  # pragma: no cover
        django_version = ""
    return {"django_version": django_version}


def process_time_travel(context):
    """Choose time_machine unless only freezegun is installed"""
    try:
        import time_machine
    except ImportError:
        try:
            import freezegun
        except ImportError:  # pragma: no cover
            pass
        else:
            return {
                "time_travel_import": "import freezegun",
                "time_travel_call": "freezegun.freeze_time",
                "time_travel_tick": ", tick=True",
            }
    return {
        "time_travel_import": "import time_machine",
        "time_travel_call": "time_machine.travel",
        "time_travel_tick": "",
    }


def import_processor(processor):
    module_path, _sep, filter_name = processor.rpartition(".")
    module = import_module(module_path)
    return getattr(module, filter_name)


def load_processors(config):
    default_processors = (
        process_django_version,
        process_django_schema,
        process_django_request,
        process_outbound_requests,
        process_sql_queries,
        process_factories,
        process_time_travel,
    )
    custom_processors = config.get("test_generation", {}).get("trace_processors", [])
    custom_processors = map(import_processor, custom_processors)
    return tuple(chain(default_processors, custom_processors))


def run_processors(processors, context):
    for processor in processors:
        processor_output = processor(context)
        if processor_output:
            context.update(processor_output)



from __future__ import annotations

import threading
import time
import traceback
import types
from itertools import starmap
from typing import Dict, List, Literal, Optional, Tuple, TypedDict

import ulid

try:
    from django.db.models import Model as DjangoModel
except ImportError:  # pragma: no cover

    class DjangoModel:  # type: ignore
        """Stub type so isinstance returns False"""


from ..serialize import frame_path
from .core import library_filter


class ExceptionFrameInfo(TypedDict):
    path: str
    co_name: str
    locals: Dict[str, object]
    expanded_locals: Dict[str, Dict[str, object]]


class RecordedException(TypedDict):
    # Usually contains one string. Last string in the list is always
    # the one indicating which exception occurred
    exception_summary: List[str]
    exception_with_traceback: List[str]
    exception_frames: List[ExceptionFrameInfo]
    frame_id: str
    bottom_exception_frame: ExceptionFrameInfo | None
    thread: str
    thread_native_id: Optional[int]
    timestamp: float
    type: Literal["exception"]


class ExceptionFilter:
    co_names: Tuple[str, ...] = ("handle_uncaught_exception",)

    def __init__(self, config, ignore_frames, include_frames) -> None:
        self.config = config
        self.ignore_frames = ignore_frames
        self.include_frames = include_frames

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        return (
            event == "call"
            and "django" in frame.f_code.co_filename
            and frame.f_code.co_name in self.co_names
        )

    def process(
        self,
        frame: types.FrameType,
        event: str,
        arg: object,
        call_frames: List[Tuple[types.FrameType, str]],
    ):
        timestamp = time.time()
        thread = threading.current_thread()
        frame_locals = frame.f_locals
        exc_type, exc_value, exc_traceback = frame_locals["exc_info"]

        recorded_exception_frames = []
        expanded_locals_for_frames = []
        frames = (frame[0] for frame in traceback.walk_tb(exc_traceback))
        for frame in frames:
            include_match = self.include_match(frame)
            if not include_match and self.ignore_match(frame):
                continue

            if include_match or not library_filter(frame):
                frame_locals = frame.f_locals

                expanded_locals = {}
                for key, value in frame_locals.items():
                    if hasattr(value, "__dict__") and isinstance(value, DjangoModel):
                        expanded_locals[key] = vars(value)

                recorded_exception_frames.append(frame)
                expanded_locals_for_frames.append(expanded_locals)

        def serialize_exception_frame(frame, expanded_locals) -> "ExceptionFrameInfo":
            return {
                "path": frame_path(frame),
                "co_name": frame.f_code.co_name,
                "locals": frame.f_locals,
                "expanded_locals": expanded_locals,
            }

        exception_with_traceback = traceback.format_exception(
            exc_type, exc_value, exc_traceback
        )

        zipped_frames = zip(recorded_exception_frames, expanded_locals_for_frames)
        exception_frames = list(starmap(serialize_exception_frame, zipped_frames))
        bottom_exception_frame = exception_frames[-1] if exception_frames else None

        recorded_exception: RecordedException = {
            "frame_id": f"frm_{ulid.new()}",
            "exception_summary": traceback.format_exception_only(exc_type, exc_value),
            "exception_with_traceback": exception_with_traceback,
            "exception_frames": exception_frames,
            "bottom_exception_frame": bottom_exception_frame,
            "thread": thread.name,
            "thread_native_id": thread.native_id,
            "timestamp": timestamp,
            "type": "exception",
        }
        return recorded_exception

    def include_match(self, frame: types.FrameType) -> bool:
        for include_filter in self.include_frames:
            if include_filter(frame, event="return", arg=None):
                return True
        return False

    def ignore_match(self, frame: types.FrameType) -> bool:
        for ignore_filter in self.ignore_frames:
            if ignore_filter(frame, event="return", arg=None):
                return True
        return False



from __future__ import annotations

import gzip
import json
import os
import sqlite3
import subprocess
import sys
from datetime import datetime, timedelta
from runpy import run_module, run_path

import click
import httpx
import msgpack

from . import django_schema
from .config import load_config
from .db import (
    TraceNotFoundError,
    create_schema_table,
    db_connection,
    delete_traces_before,
    delete_traces_by_id,
    list_traces_from_db,
    load_trace_from_db,
    save_schema,
    save_trace_in_sqlite,
    setup_db,
    vacuum_db,
)
from .profiler import KoloProfiler
from .serialize import monkeypatch_queryset_repr
from .utils import pretty_byte_size

DATETIME_FORMATS = click.DateTime(
    (
        "%Y-%m-%d",
        "%Y-%m-%dT%H:%M:%S",
        "%Y-%m-%d %H:%M:%S",
        "%Y-%m-%dT%H:%M:%S.%f",
        "%Y-%m-%d %H:%M:%S.%f",
    )
)


DJANGO_SETTINGS_ERROR = """Django settings not found.
Use `--settings` or set the `DJANGO_SETTINGS_MODULE` environment variable."""

TRACE_NOT_FOUND_ERROR = "Could not find trace_id: `{trace_id}`"

TEST_GENERATION_ERROR = """\
Test generation dependencies are not installed.

Run `pip install 'kolo[test_generation]'` to install them.
"""


def get_profiler(one_trace_per_test) -> KoloProfiler:
    config = load_config()
    db_path = setup_db()
    return KoloProfiler(
        db_path, config=config, one_trace_per_test=one_trace_per_test, source="kolo run"
    )


def profile_module(profiler: KoloProfiler, module_name: str):
    monkeypatch_queryset_repr()
    with profiler:
        run_module(module_name, run_name="__main__", alter_sys=True)


def profile_path(profiler: KoloProfiler, path: str):
    monkeypatch_queryset_repr()
    with profiler:
        run_path(path, run_name="__main__")


@click.group()
def cli():
    """Base for all kolo command line commands"""

    # Ensure the current working directory is on the path.
    # Important when running the `kolo` script installed by setuptools.
    # Not really necessary when using `python -m kolo`, but doesn't hurt.
    # Note, we use 1, not 0: https://stackoverflow.com/q/10095037
    # This probably doesn't matter for our use case, but it doesn't hurt.
    sys.path.insert(1, ".")


def python_noop_profiler(frame, event, arg):  # pragma: no cover
    pass


@cli.command(context_settings={"ignore_unknown_options": True})
@click.argument("path")
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
@click.option("--one-trace-per-test", default=False, is_flag=True)
@click.option("--noop", default=False, is_flag=True)
def run(path, args, one_trace_per_test, noop):
    """
    Profile python code using kolo.

    PATH is the path to the python file or module being profiled.
    """
    if path == "python":
        path, *args = args
        if path == "-m":
            path, *args = args
            module = True
        else:
            module = False
    elif path.endswith(".py"):
        module = False
    else:
        module = True

    # Monkeypatch sys.argv, so the profiled code doesn't get confused
    # Without this, the profiled code would see extra args it doesn't
    # know how to handle.
    sys.argv = [path, *args]

    if noop:  # pragma: no cover
        config = load_config()
        if config.get("use_rust", True):
            from ._kolo import register_noop_profiler

            register_noop_profiler()
        else:
            sys.setprofile(python_noop_profiler)

        try:
            if module:
                run_module(path, run_name="__main__", alter_sys=True)
            else:
                run_path(path, run_name="__main__")
        finally:
            sys.setprofile(None)
        return

    profiler = get_profiler(one_trace_per_test)
    try:
        if module:
            profile_module(profiler, path)
        else:
            profile_path(profiler, path)
    finally:
        if not one_trace_per_test:  # pragma: no branch
            profiler.save_request_in_db()


@cli.group()
def trace():
    """
    List, dump, delete and load traces.
    """


@cli.command()
@click.argument("path")
@click.option(
    "--created-at",
    help="Mark this trace as created at this time.",
    type=DATETIME_FORMATS,
)
def load_trace(path, created_at=None):
    """
    DEPRECATED - Use: kolo trace load.
    """
    load_trace_base(path, created_at)  # pragma: no cover


def load_trace_base(path, created_at):
    db_path = setup_db()

    try:
        with open(path, "rb") as dump:
            raw_data = dump.read()
    except FileNotFoundError:
        raise click.ClickException(f'File "{path}" not found')

    json_string: str | None = None
    msgpack_bytes: bytes | None = None

    try:
        data = msgpack.unpackb(raw_data)
    except Exception:
        try:
            json_string = raw_data.decode("utf-8")
        except UnicodeDecodeError:
            raise click.ClickException("Trace file is not valid msgpack or json")

        try:
            data = json.loads(json_string)
        except json.JSONDecodeError:
            raise click.ClickException("Trace file is not valid msgpack or json")

        msgpack_bytes = None
    else:
        msgpack_bytes = raw_data

    try:
        trace_id = data["trace_id"]
    except KeyError:
        raise click.ClickException("Trace file is missing the `trace_id`")

    try:
        save_trace_in_sqlite(
            db_path,
            trace_id,
            json_string=json_string,
            msgpack=msgpack_bytes,
            ignore_errors=False,
            created_at=created_at,
        )
    except sqlite3.IntegrityError:
        raise click.ClickException(f"Trace ID `{trace_id}` already exists")

    click.echo(f"Loaded trace {trace_id}")


@trace.command()
@click.argument("path")
@click.option(
    "--created-at",
    help="Mark this trace as created at this time.",
    type=DATETIME_FORMATS,
)
def load(path, created_at):
    """
    Load a trace from a file into the Kolo database.
    """
    load_trace_base(path, created_at)


@cli.command()
@click.argument("trace_id")
@click.option("--file", help="The name of the file to save the trace to.")
def dump_trace(trace_id, file):
    """
    DEPRECATED - Use: kolo trace dump.
    """
    dump_trace_base(trace_id, file)  # pragma: no cover


def dump_trace_base(trace_id, file):
    db_path = setup_db()

    try:
        msgpack_data, json_data = load_trace_from_db(db_path, trace_id)
    except TraceNotFoundError:
        raise click.ClickException(TRACE_NOT_FOUND_ERROR.format(trace_id=trace_id))

    if file:
        if msgpack_data is None:
            assert json_data is not None
            with open(file, "w") as f:
                f.write(json_data)
        else:
            with open(file, "wb") as f:
                f.write(msgpack_data)

    else:
        if msgpack_data is None:
            click.echo(json_data)
        else:
            click.echo(msgpack_data, nl=False)


@trace.command()
@click.argument("trace_id")
@click.option("--file", help="The name of the file to save the trace to.")
def dump(trace_id, file):
    """
    Dump a trace from the Kolo database to stdout or a specified file.
    """
    dump_trace_base(trace_id, file)


@cli.command()
def list_traces():
    """
    DEPRECATED - Use: kolo trace list.
    """
    list_traces_base()  # pragma: no cover


def list_traces_base(count=500, reverse=False):
    db_path = setup_db()

    traces = list_traces_from_db(db_path, count=count, reverse=reverse)

    if not traces:
        click.echo("No traces found")
        return

    for trace, timestamp, size in traces:
        size = pretty_byte_size(size)
        click.echo(f"{trace} at {timestamp} ({size})")


@trace.command()
@click.option("--count", help="The number of rows to show.", default=500)
@click.option(
    "--reverse",
    help="Reverse the order of the rows: newest at the bottom.",
    default=False,
    is_flag=True,
)
def list(count, reverse):
    """
    List of recent traces stored by kolo.
    """
    list_traces_base(count, reverse)


@cli.command()
@click.option(
    "--before",
    help="Delete traces older than this datetime.",
    type=DATETIME_FORMATS,
)
@click.option(
    "--vacuum",
    help="Recover disk space from the Kolo database.",
    default=False,
    is_flag=True,
)
def delete_old_traces(before, vacuum):
    """
    DEPRECATED - Use: kolo trace delete --old
    """
    delete_traces_base(
        trace_ids=(), old=True, before=before, vacuum=vacuum
    )  # pragma: no cover


def delete_traces_base(trace_ids, old, before, vacuum):
    db_path = setup_db()

    if trace_ids:
        delete_traces_by_id(db_path, trace_ids)
    elif old:
        if before is None:
            before = datetime.now() - timedelta(days=30)

        deleted_count = delete_traces_before(db_path, before)
        click.echo(f"Deleted {deleted_count} old traces created before {before}.")

    if vacuum:
        vacuum_db(db_path)


@trace.command()
@click.argument("trace_ids", required=False, nargs=-1)
@click.option("--old", is_flag=True, default=False, help="Delete old traces.")
@click.option(
    "--before",
    help="Delete traces older than this datetime. Must be used with `--old`.",
    type=DATETIME_FORMATS,
)
@click.option(
    "--vacuum",
    help="Recover disk space from the Kolo database.",
    default=False,
    is_flag=True,
)
def delete(trace_ids, old, before, vacuum):
    """
    Delete one or more traces stored by Kolo.
    """

    if before is not None and old is False:
        raise click.ClickException("--before requires --old")

    if old is False and not trace_ids and vacuum is False:
        raise click.ClickException("Must specify either TRACE_IDS, --old, or --vacuum")

    if trace_ids and old:
        raise click.ClickException("Cannot specify TRACE_IDS and --old together")

    delete_traces_base(trace_ids, old, before, vacuum)


def manage_py_settings():
    import ast

    try:
        with open("manage.py") as f:
            data = f.read()
    except OSError:  # pragma: no cover
        return None
    source = ast.parse(data, "manage.py")
    for node in ast.walk(source):
        if (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Attribute)
            and node.func.attr == "setdefault"
            and isinstance(node.args[0], ast.Constant)
            and node.args[0].value == "DJANGO_SETTINGS_MODULE"
            and isinstance(node.args[1], ast.Constant)
        ):
            return node.args[1].value  # pragma: no cover
    return None


def load_django(settings):
    import django

    if settings:
        os.environ["DJANGO_SETTINGS_MODULE"] = settings
    elif os.environ.get("DJANGO_SETTINGS_MODULE"):
        pass
    else:
        settings = manage_py_settings()
        if settings:
            os.environ["DJANGO_SETTINGS_MODULE"] = settings  # pragma: no cover
        else:
            raise click.ClickException(DJANGO_SETTINGS_ERROR)

    django.setup()


@cli.command()
@click.argument("trace_id")
@click.option("--test-name", default="test_my_view")
@click.option("--test-class", default="MyTestCase")
@click.option("--file", type=click.File("w"))
@click.option("--settings", default="")
@click.option("--template", default="")
def generate_test(trace_id, test_name, test_class, file, settings, template):
    import logging

    logging.disable()

    try:
        try:
            from .generate_tests import generate_from_trace_id
        except ModuleNotFoundError:  # pragma: no cover
            raise click.ClickException(TEST_GENERATION_ERROR)

        load_django(settings)

        try:
            test_code = generate_from_trace_id(
                trace_id, test_class, test_name, template
            )
        except TraceNotFoundError as e:
            raise click.ClickException(TRACE_NOT_FOUND_ERROR.format(trace_id=e.args[0]))

        if file:
            file.write(test_code)
        else:
            click.echo(test_code)
    finally:
        logging.disable(logging.NOTSET)


@cli.command()
@click.option("--settings", default="")
def store_django_model_schema(settings):
    from .git import COMMIT_SHA

    load_django(settings)

    schema = django_schema.get_schema()

    db_path = setup_db()
    with db_connection(db_path) as connection:
        create_schema_table(connection)
        save_schema(connection, schema, COMMIT_SHA)


@cli.command()
def dbshell():  # pragma: no cover
    """
    Open a sqlite3 shell to the Kolo database.
    """
    db_path = setup_db()
    subprocess.run(["sqlite3", db_path], check=True)


@trace.command()
@click.argument("trace_id")
def upload(trace_id):
    """
    Upload a trace to the kolo dashboard
    """
    db_path = setup_db()

    try:
        msgpack_data, json_data = load_trace_from_db(db_path, trace_id)
    except TraceNotFoundError:
        raise click.ClickException(TRACE_NOT_FOUND_ERROR.format(trace_id=trace_id))

    base_url = os.environ.get("KOLO_BASE_URL", "https://my.kolo.app")
    url = f"{base_url}/api/traces/"
    if msgpack_data is None:
        assert json_data is not None
        payload = gzip.compress(json_data.encode("utf-8"))
    else:
        payload = gzip.compress(msgpack_data)
    response = httpx.post(url, files={"data": payload})
    if response.status_code == 201:
        click.echo(f"{trace_id} uploaded successfully!")
    else:
        errors = response.json()
        raise click.ClickException(errors)


@trace.command()
@click.argument("trace_id")
def download(trace_id):
    """
    Download a trace from the kolo dashboard
    """
    db_path = setup_db()

    base_url = os.environ.get("KOLO_BASE_URL", "https://my.kolo.app")
    url = f"{base_url}/api/traces/{trace_id}/"
    response = httpx.get(url)

    if response.status_code == 404:
        raise click.ClickException(f"`{trace_id}` was not found by the server.")
    elif response.status_code != 200:
        raise click.ClickException(f"Unexpected status code: {response.status_code}.")

    try:
        raw_data = gzip.decompress(response.content)
    except gzip.BadGzipFile:
        raise click.ClickException("Downloaded trace was not gzipped.")

    try:
        msgpack.unpackb(raw_data, strict_map_key=False)
    except Exception:
        raise click.ClickException("Downloaded trace was not valid msgpack.")

    try:
        save_trace_in_sqlite(
            db_path,
            trace_id,
            msgpack=raw_data,
            ignore_errors=False,
        )
    except sqlite3.IntegrityError:
        raise click.ClickException(f"`{trace_id}` already exists.")

    click.echo(f"`{trace_id}` downloaded successfully!")


if __name__ == "__main__":
    cli()  # pragma: no cover
